# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Segmentation models.
"""

import os
import warnings
from zipfile import ZipFile

import bigfish.stack as stack

import tensorflow as tf

from tensorflow.python.keras.layers import Input
from tensorflow.python.keras.layers import Softmax
from tensorflow.python.keras.layers import Concatenate
from tensorflow.python.keras.engine.training import Model

from .utils_models import EncoderDecoder
from .utils_models import SameConv


# ### Pre-trained models ###

def load_pretrained_model(channel, model_name):
    """Build and compile a model, then load its pretrained weights.

    Parameters
    ----------
    channel : str
        Input channel for the model ('nuc', 'cell' or 'double').
    model_name : str
        Name of the model used ('3-classes' or 'double_distance_edge).

    Returns
    -------
    model : tensorflow.keras.model object
        Pretrained Unet model.

    """
    # TODO fix warning partial restoration with distance model

    # check parameters
    stack.check_parameter(channel=str,
                          model_name=str)

    # unet 3-classes for nucleus segmentation
    if model_name == "3_classes" and channel == "nuc":
        model = build_compile_3_classes_model()

    # unet distance map to edge for both nucleus and cell segmentation
    elif model_name == "distance_edge" and channel == "double":
        model = build_compile_double_distance_model()

    else:
        raise ValueError("Model name and channel to segment are not "
                         "consistent: {0} - {1}.".format(channel, model_name))

    # load weights
    path_pretrained_directory = check_pretrained_weights(channel, model_name)
    path_checkpoint = os.path.join(
        path_pretrained_directory, "checkpoint")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        model.load_weights(path_checkpoint)

    return model


def check_pretrained_weights(channel, model_name):
    """Check pretrained weights exist and download them if necessary.

    Parameters
    ----------
    channel : str
        Input channel for the model ('nuc', 'cell' or 'double').
    model_name : str
        Name of the model used ('3-classes' or 'distance_edge').

    Returns
    -------
    directory_checkpoint : str
        Path of the checkpoint directory.

    """
    # check parameters
    stack.check_parameter(channel=str,
                          model_name=str)

    # get path checkpoint
    path_weights_directory = _get_weights_directory()
    pretrained_directory = "_".join([channel, model_name])
    path_directory = os.path.join(path_weights_directory, pretrained_directory)
    download = False

    # get url and hash
    if channel == "nuc" and model_name == "3_classes":
        url_zip_file = "https://github.com/fish-quant/big-fish-examples/releases/download/weights_1/nuc_3_classes.zip"
        hash_checkpoint = "02988027faf3f16b4088ee83c2ade14098e8ffb325c23a576cc639dae48aa936"
        hash_data = "284d6b8cadb3eddd691f1407bfdd1a7e6fa085c57a2feac07446e84aa3b1baf8"
        hash_index = "c924f0f2be179340dc5c75e21833fd831d5a4efdfb48382edbbcff748a162bc4"
        hash_log = "b6d08b9fbacd3430d89eef2cd5c3246eb8c93e7b970ba17a76c8a9d72f58c8e7"
    elif channel == "double" and model_name == "distance_edge":
        url_zip_file = "https://github.com/fish-quant/big-fish-examples/releases/download/weights_0/double_distance_edge.zip"
        hash_checkpoint = "02988027faf3f16b4088ee83c2ade14098e8ffb325c23a576cc639dae48aa936"
        hash_data = "528aecbc6418df8d0d73fb23b3518ef779ae5135e15ca3e2e67518726f998d4d"
        hash_index = "cbebdd86868e46507733fccca589f940a438f4fde6f5d935e61f4112047db63e"
        hash_log = "63bc7e629d464bbf6cd3afd466f4d89a3be10edb2633de7ab825384a61326a09"
    else:
        raise ValueError("Model name and channel to segment are not "
                         "consistent: {0} - {1}.".format(channel, model_name))

    # case where pretrained directory exists
    if os.path.isdir(path_directory):

        # paths
        path_checkpoint = os.path.join(
            path_directory, "checkpoint")
        path_data = os.path.join(
            path_directory, "checkpoint.data-00000-of-00001")
        path_index = os.path.join(
            path_directory, "checkpoint.index")
        path_log = os.path.join(
            path_directory, "log")

        # checkpoint available and not corrupted
        if os.path.exists(path_checkpoint):
            try:
                stack.check_hash(path_checkpoint, hash_checkpoint)
            except IOError:
                print("{0} seems corrupted.".format(path_checkpoint))
                download = True
        else:
            download = True

        # data available and not corrupted
        if os.path.exists(path_data):
            try:
                stack.check_hash(path_data, hash_data)
            except IOError:
                print("{0} seems corrupted.".format(path_data))
                download = True
        else:
            download = True

        # index available and not corrupted
        if os.path.exists(path_index):
            try:
                stack.check_hash(path_index, hash_index)
            except IOError:
                print("{0} seems corrupted.".format(path_index))
                download = True
        else:
            download = True

        # log available and not corrupted
        if os.path.exists(path_log):
            try:
                stack.check_hash(path_log, hash_log)
            except IOError:
                print("{0} seems corrupted.".format(path_log))
                download = True
        else:
            download = True

    # case where pretrained directory does not exist
    else:
        download = True

    # download checkpoint files
    if download:
        print("downloading model weights...")

        # download zipfile
        stack.load_and_save_url(
            remote_url=url_zip_file,
            directory=path_weights_directory)

        # unzip
        path_zipfile = os.path.join(
            path_weights_directory, "{0}.zip".format(pretrained_directory))
        with ZipFile(path_zipfile, 'r') as zip:
            zip.extract(
                member="{0}/checkpoint".format(pretrained_directory),
                path=path_weights_directory)
            zip.extract(
                member="{0}/checkpoint.data-00000-of-00001"
                    .format(pretrained_directory),
                path=path_weights_directory)
            zip.extract(
                member="{0}/checkpoint.index".format(pretrained_directory),
                path=path_weights_directory)
            zip.extract(
                member="{0}/log".format(pretrained_directory),
                path=path_weights_directory)

        # check files consistency
        path = os.path.join(
            path_weights_directory,
            "{0}/checkpoint".format(pretrained_directory))
        stack.check_hash(path, hash_checkpoint)
        path = os.path.join(
            path_weights_directory,
            "{0}/checkpoint.data-00000-of-00001".format(pretrained_directory))
        stack.check_hash(path, hash_data)
        path = os.path.join(
            path_weights_directory,
            "{0}/checkpoint.index".format(pretrained_directory))
        stack.check_hash(path, hash_index)
        path = os.path.join(
            path_weights_directory,
            "{0}/log".format(pretrained_directory))
        stack.check_hash(path, hash_log)

        # remove zipfile
        os.remove(path_zipfile)

    return path_directory


def _get_weights_directory():
    path = os.path.realpath(__file__)
    path = os.path.dirname(path)
    path = os.path.dirname(path)
    path = os.path.join(path, "weights")

    return path


# ### Models 3-classes ###

def build_compile_3_classes_model():
    """Build and compile a Unet model to predict 3 classes from nucleus or
    cell images: background, edge and foreground.

    Returns
    -------
    model_3_classes : tensorflow.keras.model object
        Compiled Unet model.

    """
    # define inputs
    inputs_image = Input(
        shape=(None, None, 1), dtype="float32", name="image")

    # define model
    outputs = _get_3_classes_model(inputs_image)
    model_3_classes = Model(
        inputs_image,
        outputs,
        name="3ClassesModel")

    # losses
    loss = tf.keras.losses.SparseCategoricalCrossentropy()

    # metrics
    accuracy = tf.metrics.SparseCategoricalAccuracy(name="accuracy")

    # compile model
    model_3_classes.compile(
        optimizer='adam',
        loss=loss,
        metrics=accuracy)

    return model_3_classes


def _get_3_classes_model(inputs):
    """Build Unet architecture that return a 3-classes prediction.

    Parameters
    ----------
    inputs : tensorflow.keras.Input object
        Input layer with shape (B, H, W, 1) or (H, W, 1).

    Returns
    -------
    output : tensorflow.keras.layers object
        Output layer (softmax) with shape (B, H, W, 3) or (H, W, 3).

    """
    # compute feature map
    features_core = EncoderDecoder(
        name="encoder_decoder")(inputs)  # (B, H, W, 32)

    # compute 3-classes output
    features_3_classes = SameConv(
        filters=3,
        kernel_size=(1, 1),
        activation="linear",
        name="final_conv")(features_core)  # (B, H, W, 3)
    output = Softmax(
        axis=-1,
        name="label_3")(features_3_classes)  # (B, H, W, 3)

    return output


# ### Models distance map ###

def build_compile_distance_model():
    """Build and compile a Unet model to predict foreground and a distance map
    from nucleus or cell images.

    Returns
    -------
    model_distance : Tensorflow model
        Compiled Unet model.

    """
    # define inputs
    inputs_image = Input(
        shape=(None, None, 1), dtype="float32", name="image")

    # define model
    output_surface, output_distance = _get_distance_model(inputs_image)
    outputs = [output_surface, output_distance]
    model_distance = Model(
        inputs_image,
        outputs,
        name="DistanceModel")

    # losses
    loss_surface = tf.keras.losses.BinaryCrossentropy()
    loss_distance = tf.keras.losses.MeanAbsoluteError()
    losses = [[loss_surface], [loss_distance]]
    losses_weight = [[1.0], [1.0]]

    # metrics
    metric_surface = tf.metrics.BinaryAccuracy(name="accuracy")
    metric_distance = tf.metrics.MeanAbsoluteError(name="mae")
    metrics = [[metric_surface], [metric_distance]]

    # compile model
    model_distance.compile(
        optimizer='adam',
        loss=losses,
        loss_weights=losses_weight,
        metrics=metrics)

    return model_distance


def _get_distance_model(inputs):
    """Build Unet architecture that return a distance map.

    Parameters
    ----------
    inputs : tensorflow.keras.Input object
        Input layer with shape (B, H, W, 1) or (H, W, 1).

    Returns
    -------
    output_surface : tensorflow.keras.layers object
        Output layer for the foreground/background prediction with shape
        (B, H, W, 1) or (H, W, 1).
    output_distance : tensorflow.keras.layers object
        Output layer for the distance map with shape (B, H, W, 1) or (H, W, 1).

    """
    # compute feature map
    features_core = EncoderDecoder(
        name="encoder_decoder")(inputs)  # (B, H, W, 32)

    # compute surface output
    output_surface = SameConv(
        filters=1,
        kernel_size=(1, 1),
        activation="sigmoid",
        name="label_2")(features_core)  # (B, H, W, 1)

    # compute distance output
    output_distance = SameConv(
        filters=1,
        kernel_size=(1, 1),
        activation="relu",
        name="label_distance")(features_core)  # (B, H, W, 1)

    return output_surface, output_distance


def build_compile_double_distance_model():
    """Build and compile a Unet model to predict foreground and a distance map
    from nucleus and cell images.

    This model version takes two images as input (for nucleus and cell).

    Returns
    -------
    model_distance : Tensorflow model
        Compiled Unet model.

    """
    # define inputs
    inputs_nuc = Input(
        shape=(None, None, 1), dtype="float32", name="nuc")
    inputs_cell = Input(
        shape=(None, None, 1), dtype="float32", name="cell")
    inputs = [inputs_nuc, inputs_cell]

    # define model
    (output_distance_nuc, output_surface_cell,
     output_distance_cell) = _get_double_distance_model(inputs)
    outputs = [output_distance_nuc, output_surface_cell, output_distance_cell]
    model_distance = Model(
        inputs,
        outputs,
        name="DoubleDistanceModel")

    # losses
    loss_distance_nuc = tf.keras.losses.MeanAbsoluteError()
    loss_surface_cell = tf.keras.losses.BinaryCrossentropy()
    loss_distance_cell = tf.keras.losses.MeanAbsoluteError()
    losses = [[loss_distance_nuc],
              [loss_surface_cell], [loss_distance_cell]]
    losses_weight = [[1.0], [1.0], [1.0]]

    # metrics
    metric_distance_nuc = tf.metrics.MeanAbsoluteError(name="mae")
    metric_surface_cell = tf.metrics.BinaryAccuracy(name="accuracy")
    metric_distance_cell = tf.metrics.MeanAbsoluteError(name="mae")
    metrics = [[metric_distance_nuc],
               [metric_surface_cell], [metric_distance_cell]]

    # compile model
    model_distance.compile(
        optimizer='adam',
        loss=losses,
        loss_weights=losses_weight,
        metrics=metrics)

    return model_distance


def _get_double_distance_model(inputs):
    """Build Unet architecture that return nucleus and cell distance maps, and
    a cell surface prediction.

    Parameters
    ----------
    inputs : List[tensorflow.keras.Input object]
        List of two input layer with shape (B, H, W, 1) or (H, W, 1).

    Returns
    -------
    output_distance_nuc : tensorflow.keras.layers object
        Output layer for the nucleus distance map with shape (B, H, W, 1) or
        (H, W, 1).
    output_surface_cell : tensorflow.keras.layers object
        Output layer for the cell foreground/background prediction with shape
        (B, H, W, 1) or (H, W, 1).
    output_distance_cell : tensorflow.keras.layers object
        Output layer for the cell distance map with shape (B, H, W, 1) or
        (H, W, 1).

    """
    # compute feature map
    inputs_nuc, inputs_cell = inputs
    inputs = Concatenate(
        axis=-1)([inputs_nuc, inputs_cell])  # (B, H, W, 2)
    features_core = EncoderDecoder(
        name="encoder_decoder")(inputs)  # (B, H, W, 32)

    # compute distance output nucleus
    output_distance_nuc = SameConv(
        filters=1,
        kernel_size=(1, 1),
        activation="relu",
        name="label_distance_nuc")(features_core)  # (B, H, W, 1)

    # compute surface output cell
    output_surface_cell = SameConv(
        filters=1,
        kernel_size=(1, 1),
        activation="sigmoid",
        name="label_2_cell")(features_core)  # (B, H, W, 1)

    # compute distance output cell
    output_distance_cell = SameConv(
        filters=1,
        kernel_size=(1, 1),
        activation="relu",
        name="label_distance_cell")(features_core)  # (B, H, W, 1)

    return output_distance_nuc, output_surface_cell, output_distance_cell



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
The bigfish.plot subpackage includes functions to plot images and results.
"""

from .plot_images import plot_yx
from .plot_images import plot_images
from .plot_images import plot_segmentation
from .plot_images import plot_segmentation_boundary
from .plot_images import plot_segmentation_diff

from .plot_images import plot_detection
from .plot_images import plot_reference_spot
from .plot_images import plot_cell

from .plot_quality import plot_sharpness
from .plot_quality import plot_elbow

from .utils import save_plot
from .utils import get_minmax_values
from .utils import create_colormap


_images = [
    "plot_yx",
    "plot_images",
    "plot_segmentation",
    "plot_segmentation_boundary",
    "plot_segmentation_diff",
    "plot_detection",
    "plot_reference_spot",
    "plot_cell"]

_quality = [
    "plot_sharpness",
    "plot_elbow"]

_utils = [
    "save_plot",
    "get_minmax_values",
    "create_colormap"]

__all__ = _images + _quality + _utils



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for bigfish.stack.io module.
"""

import os
import pytest
import mrc
import tempfile

import numpy as np
import pandas as pd
import bigfish.stack as stack

from numpy.testing import assert_array_equal

# TODO test bigfish.stack.read_cell_extracted
# TODO test bigfish.stack.save_cell_extracted


@pytest.mark.parametrize("shape", [
    (8, 8), (8, 8, 8), (8, 8, 8, 8), (8, 8, 8, 8, 8)])
@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float16, np.float32, np.float64, bool])
@pytest.mark.parametrize("extension", [
    "png", "jpg", "jpeg", "tif", "tiff"])
def test_image(shape, dtype, extension):
    # build a temporary directory and save tensors inside
    with tempfile.TemporaryDirectory() as tmp_dir:
        test = np.zeros(shape, dtype=dtype)
        path = os.path.join(tmp_dir, "test." + extension)

        # error: boolean multidimensional image
        if (extension in ["png", "jpg", "jpeg", "tif", "tiff"]
                and len(test.shape) > 2
                and test.dtype == bool):
            with pytest.raises(ValueError):
                stack.save_image(test, path)

        # error: non-boolean multidimensional image with 'png', 'jpg' or 'jpeg'
        elif (extension in ["png", "jpg", "jpeg"]
                and len(test.shape) > 2
                and test.dtype != bool):
            with pytest.raises(ValueError):
                stack.save_image(test, path)

        # error: boolean 2-d image with 'tig' and 'tiff'
        elif (extension in ["tif", "tiff"]
                and len(test.shape) == 2
                and test.dtype == bool):
            with pytest.raises(ValueError):
                stack.save_image(test, path)

        # warning: 2-d image with 'png', 'jpg' or 'jpeg'
        elif (extension in ["png", "jpg", "jpeg"]
                and len(test.shape) == 2):
            with pytest.warns(UserWarning):
                stack.save_image(test, path)
                tensor = stack.read_image(path, sanity_check=True)
                assert_array_equal(test, tensor)

        # others valid images
        else:
            stack.save_image(test, path)
            tensor = stack.read_image(path, sanity_check=True)
            assert_array_equal(test, tensor)
            assert test.dtype == tensor.dtype


def test_image_specific():
    # build a temporary directory and save tensors inside
    with tempfile.TemporaryDirectory() as tmp_dir:
        # non-supported image (1 dimension)
        test = np.array([1, 2, 3], dtype=np.float32)
        path = os.path.join(tmp_dir, "test")
        with pytest.raises(ValueError):
            stack.save_image(test, path)

        # non-supported image (6 dimensions)
        test = np.zeros((8, 8, 8, 8, 8, 8), dtype=np.float32)
        path = os.path.join(tmp_dir, "test")
        with pytest.raises(ValueError):
            stack.save_image(test, path)

        # non-supported image (missing values)
        test = np.zeros((8, 8), dtype=np.float32)
        test[0, 0] = np.nan
        path = os.path.join(tmp_dir, "test")
        with pytest.raises(ValueError):
            stack.save_image(test, path)

        # non-supported extensions
        test = np.zeros((8, 8), dtype=np.float32)
        path = os.path.join(tmp_dir, "test.foo")
        with pytest.raises(ValueError):
            stack.save_image(test, path)
        path = os.path.join(tmp_dir, "test")
        with pytest.raises(ValueError):
            stack.save_image(test, path, extension="bar")


@pytest.mark.parametrize("dtype", [np.uint16, np.int16, np.int32, np.float32])
def test_dv(dtype):
    # build a temporary directory and save a dv file inside
    with tempfile.TemporaryDirectory() as tmp_dir:
        test = np.zeros((8, 8, 8, 8), dtype=dtype)
        path = os.path.join(tmp_dir, "test")
        mrc.imsave(path, test)

        # read and test the dv file
        path = os.path.join(tmp_dir, "test")
        tensor = stack.read_dv(path, sanity_check=True)
        assert_array_equal(test, tensor)
        assert test.dtype == tensor.dtype


@pytest.mark.parametrize("shape", [
    (8, 8), (8, 8, 8), (8, 8, 8, 8), (8, 8, 8, 8, 8)])
@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float16, np.float32, np.float64, bool])
def test_npy(shape, dtype):
    # build a temporary directory and save tensors inside
    with tempfile.TemporaryDirectory() as tmp_dir:
        test = np.zeros(shape, dtype=dtype)
        path = os.path.join(tmp_dir, "test.npy")
        stack.save_array(test, path)
        tensor = stack.read_array(path)
        assert_array_equal(test, tensor)
        assert test.dtype == tensor.dtype


@pytest.mark.parametrize("shape", [
    (8, 8), (8, 8, 8), (8, 8, 8, 8), (8, 8, 8, 8, 8)])
@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float16, np.float32, np.float64, bool])
def test_npz(shape, dtype):
    # build a temporary directory and save tensors inside
    with tempfile.TemporaryDirectory() as tmp_dir:
        test_1 = np.zeros(shape, dtype=dtype)
        test_2 = np.ones(shape, dtype=dtype)
        path = os.path.join(tmp_dir, "test.npz")
        np.savez(path, test_1=test_1, test_2=test_2)
        data = stack.read_uncompressed(path)
        assert data.files == ["test_1", "test_2"]
        assert_array_equal(test_1, data["test_1"])
        assert_array_equal(test_2, data["test_2"])
        assert test_1.dtype == data["test_1"].dtype
        assert test_2.dtype == data["test_2"].dtype


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float16, np.float32, np.float64])
@pytest.mark.parametrize("delimiter", [";", ",", ":", "\t"])
def test_csv_numpy(dtype, delimiter):
    # build a temporary directory
    with tempfile.TemporaryDirectory() as tmp_dir:
        # save arrays in csv file
        test_array = np.zeros((10, 2), dtype=dtype)
        path = os.path.join(tmp_dir, "test_array.csv")
        stack.save_data_to_csv(test_array, path, delimiter=delimiter)

        # read csv file
        array = stack.read_array_from_csv(path, delimiter=delimiter)
        assert array.dtype == np.float64
        array = stack.read_array_from_csv(path, dtype, delimiter)
        assert_array_equal(test_array, array)
        assert array.dtype == dtype


@pytest.mark.parametrize("delimiter", [";", ",", ":", "\t"])
def test_csv_pandas(delimiter):
    # build a temporary directory
    with tempfile.TemporaryDirectory() as tmp_dir:
        # save pandas objects in csv file
        test_series = pd.Series([0.1, 2.5, 3.7], name="a")
        test_dataframe = pd.DataFrame({"a": [0, 2, 3],
                                       "b": [0.1, 2.5, 3.7],
                                       "c": ["dog", "cat", "bird"]})
        path = os.path.join(tmp_dir, "test_series.csv")
        stack.save_data_to_csv(test_series, path, delimiter=delimiter)
        path = os.path.join(tmp_dir, "test_dataframe.csv")
        stack.save_data_to_csv(test_dataframe, path, delimiter=delimiter)

        # read csv files
        path = os.path.join(tmp_dir, "test_series.csv")
        df = stack.read_dataframe_from_csv(path, delimiter=delimiter)
        pd.testing.assert_frame_equal(test_series.to_frame(), df)
        path = os.path.join(tmp_dir, "test_dataframe.csv")
        df = stack.read_dataframe_from_csv(path, delimiter=delimiter)
        pd.testing.assert_frame_equal(test_dataframe, df)



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Functions to craft features.
"""

import warnings
import numpy as np
from scipy import ndimage as ndi

from skimage.morphology import binary_opening
from skimage.morphology.selem import disk

import bigfish.stack as stack
from .input_preparation import prepare_extracted_data


# ### Main functions ###

def compute_features(cell_mask, nuc_mask, ndim, rna_coord, smfish=None,
                     voxel_size_yx=None, foci_coord=None,
                     centrosome_coord=None,
                     compute_distance=False, compute_intranuclear=False,
                     compute_protrusion=False, compute_dispersion=False,
                     compute_topography=False, compute_foci=False,
                     compute_area=False, compute_centrosome=False,
                     return_names=False):
    """Compute requested features.

    Parameters
    ----------
    cell_mask : np.ndarray, np.uint, np.int or bool
        Surface of the cell with shape (y, x).
    nuc_mask: np.ndarray, np.uint, np.int or bool
        Surface of the nucleus with shape (y, x).
    ndim : int
        Number of spatial dimensions to consider (2 or 3).
    rna_coord : np.ndarray, np.int64
        Coordinates of the detected spots with shape (nb_spots, 4) or
        (nb_spots, 3). One coordinate per dimension (zyx or yx dimensions)
        plus the index of the cluster assigned to the spot. If no cluster was
        assigned, value is -1. If cluster id is not provided foci related
        features are not computed.
    smfish : np.ndarray, np.uint
        Image of RNAs, with shape (y, x).
    voxel_size_yx : int, float or None
        Size of a voxel on the yx plan, in nanometer.
    foci_coord : np.ndarray, np.int64
        Array with shape (nb_foci, 5) or (nb_foci, 4). One coordinate per
        dimension for the foci centroid (zyx or yx coordinates), the number of
        spots detected in the foci and its index.
    centrosome_coord : np.ndarray, np.int64
        Coordinates of the detected centrosome with shape (nb_elements, 3) or
        (nb_elements, 2). One coordinate per dimension (zyx or yx dimensions).
        These coordinates are mandatory to compute centrosome related features.
    compute_distance : bool
        Compute distance related features.
    compute_intranuclear : bool
        Compute nucleus related features.
    compute_protrusion : bool
        Compute protrusion related features.
    compute_dispersion : bool
        Compute dispersion indices.
    compute_topography : bool
        Compute topographic features.
    compute_foci : bool
        Compute foci related features.
    compute_area : bool
        Compute area related features.
    compute_centrosome : bool
        Compute centrosome related features.
    return_names : bool
        Return features names.

    Returns
    -------
    features : np.ndarray, np.float32
        Array of features.

    """
    # check parameters
    stack.check_parameter(voxel_size_yx=(int, float, type(None)),
                          compute_distance=bool,
                          compute_intranuclear=bool,
                          compute_protrusion=bool,
                          compute_dispersion=bool,
                          compute_topography=bool,
                          compute_foci=bool,
                          compute_area=bool,
                          compute_centrosome=bool,
                          return_names=bool)
    if smfish is not None:
        stack.check_array(smfish, ndim=[2, 3], dtype=[np.uint8, np.uint16])
        if smfish.ndim == 3:
            smfish = stack.maximum_projection(smfish)
    if foci_coord is not None:
        stack.check_array(foci_coord, ndim=2, dtype=np.int64)

    # prepare input data
    (cell_mask,
     distance_cell, distance_cell_normalized,
     centroid_cell, distance_centroid_cell,
     nuc_mask, cell_mask_out_nuc,
     distance_nuc, distance_nuc_normalized,
     centroid_nuc, distance_centroid_nuc,
     rna_coord_out_nuc,
     centroid_rna, distance_centroid_rna,
     centroid_rna_out_nuc, distance_centroid_rna_out_nuc,
     distance_centrosome) = prepare_extracted_data(
        cell_mask, nuc_mask, ndim, rna_coord, centrosome_coord)

    # initialization
    features = ()
    names_features_distance = False
    names_features_intranuclear = False
    names_features_protrusion = False
    names_features_dispersion = False
    names_features_topography = False
    names_features_foci = False
    names_features_area = False
    names_features_centrosome = False

    # distance related features
    if compute_distance:
        features += features_distance(
            rna_coord, distance_cell, distance_nuc, cell_mask, ndim, False)
        names_features_distance = True

    # nucleus related features
    if compute_intranuclear:
        features += features_in_out_nucleus(
            rna_coord, rna_coord_out_nuc, False)
        names_features_intranuclear = True

    # protrusion related features
    if compute_protrusion:
        features += features_protrusion(
            rna_coord, cell_mask, nuc_mask, ndim, voxel_size_yx, False)
        names_features_protrusion = True

    # dispersion indices
    if compute_dispersion and smfish is not None:
        features += features_dispersion(
            smfish, rna_coord, centroid_rna, cell_mask, centroid_cell,
            centroid_nuc, ndim, False)
        names_features_dispersion = True
    elif compute_dispersion and smfish is None:
        raise ValueError("Dispersion features can't be computed because "
                         "'smfish' is not provided.")

    # topographic features
    if compute_topography and voxel_size_yx is not None:
        features += features_topography(
            rna_coord, cell_mask, nuc_mask, cell_mask_out_nuc, ndim,
            voxel_size_yx, False)
        names_features_topography = True
    elif compute_topography and voxel_size_yx is None:
        raise ValueError("Topographic features can't be computed because "
                         "'voxel_size_yx' is not provided.")

    # foci related features
    if compute_foci and foci_coord is not None:
        features += features_foci(
            rna_coord, foci_coord, ndim, False)
        names_features_foci = True
    elif compute_foci and foci_coord is None:
        raise ValueError("Foci related features can't be computed because "
                         "'foci_coord' is not provided.")

    # area related features
    if compute_area:
        features += features_area(
            cell_mask, nuc_mask, cell_mask_out_nuc, False)
        names_features_area = True

    # centrosome related features
    if (compute_centrosome and centrosome_coord is not None
            and voxel_size_yx is not None and smfish is not None):
        features += features_centrosome(
            smfish, rna_coord, distance_centrosome, cell_mask, ndim,
            voxel_size_yx, False)
        names_features_centrosome = True
    elif compute_centrosome and centrosome_coord is None:
        raise ValueError("Centrosome related features can't be computed "
                         "because 'centrosome_coord' is not provided.")
    elif compute_centrosome and voxel_size_yx is None:
        raise ValueError("Centrosome related features can't be computed "
                         "because 'voxel_size_yx' is not provided.")
    elif compute_centrosome and smfish is None:
        raise ValueError("Centrosome related features can't be computed "
                         "because 'smfish' is not provided.")

    # format features
    features = np.array(features, dtype=np.float32)
    features = np.round(features, decimals=2)

    if return_names:
        features_names = get_features_name(
            names_features_distance=names_features_distance,
            names_features_intranuclear=names_features_intranuclear,
            names_features_protrusion=names_features_protrusion,
            names_features_dispersion=names_features_dispersion,
            names_features_topography=names_features_topography,
            names_features_foci=names_features_foci,
            names_features_area=names_features_area,
            names_features_centrosome=names_features_centrosome)
        return features, features_names

    return features


def get_features_name(names_features_distance=False,
                      names_features_intranuclear=False,
                      names_features_protrusion=False,
                      names_features_dispersion=False,
                      names_features_topography=False,
                      names_features_foci=False,
                      names_features_area=False,
                      names_features_centrosome=False):
    """Return the current list of features names.

    Parameters
    ----------
    names_features_distance : bool
        Return names of features related to distances from nucleus or cell
        membrane.
    names_features_intranuclear : bool
        Return names of features related to nucleus.
    names_features_protrusion : bool
        Return names of features related to protrusions.
    names_features_dispersion : bool
        Return names of features used to quantify mRNAs dispersion within the
        cell.
    names_features_topography : bool
        Return names of topographic features of the cell.
    names_features_foci : bool
        Return names of features related to foci.
    names_features_area : bool
        Return names of features related to area of the cell.
    names_features_centrosome : bool
        Return names of features related to centrosome.

    Returns
    -------
    features_name : List[str]
        A list of features name.

    """
    # check parameters
    stack.check_parameter(names_features_distance=bool,
                          names_features_intranuclear=bool,
                          names_features_protrusion=bool,
                          names_features_dispersion=bool,
                          names_features_topography=bool,
                          names_features_foci=bool,
                          names_features_area=bool,
                          names_features_centrosome=bool)

    # initialization
    features_name = []

    # get feature names
    if names_features_distance:
        features_name += ["index_mean_distance_cell",
                          "index_median_distance_cell",
                          "index_mean_distance_nuc",
                          "index_median_distance_nuc"]

    if names_features_intranuclear:
        features_name += ["proportion_rna_in_nuc",
                          "nb_rna_out_nuc",
                          "nb_rna_in_nuc"]

    if names_features_protrusion:
        features_name += ["index_rna_protrusion",
                          "proportion_rna_protrusion",
                          "protrusion_area"]

    if names_features_dispersion:
        features_name += ["index_polarization",
                          "index_dispersion",
                          "index_peripheral_distribution"]

    if names_features_topography:
        features_name += ["index_rna_nuc_edge",
                          "proportion_rna_nuc_edge"]

        a = 500
        for b in range(1000, 3001, 500):
            features_name += ["index_rna_nuc_radius_{}_{}".format(a, b),
                              "proportion_rna_nuc_radius_{}_{}".format(a, b)]
            a = b

        a = 0
        for b in range(500, 3001, 500):
            features_name += ["index_rna_cell_radius_{}_{}".format(a, b),
                              "proportion_rna_cell_radius_{}_{}".format(a, b)]
            a = b

    if names_features_foci:
        features_name += ["proportion_rna_in_foci"]

    if names_features_area:
        features_name += ["proportion_nuc_area",
                          "cell_area",
                          "nuc_area",
                          "cell_area_out_nuc"]

    if names_features_centrosome:
        features_name += ["index_mean_distance_centrosome",
                          "index_median_distance_centrosome",
                          "index_rna_centrosome",
                          "proportion_rna_centrosome",
                          "index_centrosome_dispersion"]

    return features_name


# ### Features functions ###

def features_distance(rna_coord, distance_cell, distance_nuc, cell_mask, ndim,
                      check_input=True):
    """Compute distance related features.

    Parameters
    ----------
    rna_coord : np.ndarray, np.int64
        Coordinates of the detected RNAs with zyx or yx coordinates in the
        first 3 or 2 columns.
    distance_cell : np.ndarray, np.float32
        Distance map from the cell with shape (y, x).
    distance_nuc : np.ndarray, np.float32
        Distance map from the nucleus with shape (y, x).
    cell_mask : np.ndarray, bool
        Surface of the cell with shape (y, x).
    ndim : int
        Number of spatial dimensions to consider.
    check_input : bool
        Check input validity.

    Returns
    -------
    index_mean_dist_cell : float
        Normalized mean distance of RNAs to the cell membrane.
    index_median_dist_cell : float
        Normalized median distance of RNAs to the cell membrane.
    index_mean_dist_nuc : float
        Normalized mean distance of RNAs to the nucleus.
    index_median_dist_nuc : float
        Normalized median distance of RNAs to the nucleus.

    """
    # check parameters
    stack.check_parameter(check_input=bool)
    if check_input:
        stack.check_parameter(ndim=int)
        if ndim not in [2, 3]:
            raise ValueError("'ndim' should be 2 or 3, not {0}.".format(ndim))
        stack.check_array(rna_coord, ndim=2, dtype=np.int64)
        stack.check_array(distance_cell, ndim=2,
                          dtype=[np.float16, np.float32, np.float64])
        stack.check_array(distance_nuc, ndim=2,
                          dtype=[np.float16, np.float32, np.float64])
        stack.check_array(cell_mask, ndim=2, dtype=bool)

    # case where no mRNAs are detected
    if len(rna_coord) == 0:
        features = (1., 1., 1., 1.)
        return features

    # compute mean and median distance to cell membrane
    rna_distance_cell = distance_cell[rna_coord[:, ndim - 2],
                                      rna_coord[:, ndim - 1]]
    expected_distance = np.mean(distance_cell[cell_mask])
    index_mean_dist_cell = np.mean(rna_distance_cell) / expected_distance
    expected_distance = np.median(distance_cell[cell_mask])
    index_median_dist_cell = np.median(rna_distance_cell) / expected_distance

    features = (index_mean_dist_cell, index_median_dist_cell)

    # compute mean and median distance to nucleus
    rna_distance_nuc = distance_nuc[rna_coord[:, ndim - 2],
                                    rna_coord[:, ndim - 1]]
    expected_distance = np.mean(distance_nuc[cell_mask])
    index_mean_dist_nuc = np.mean(rna_distance_nuc) / expected_distance
    expected_distance = np.median(distance_nuc[cell_mask])
    index_median_dist_nuc = np.median(rna_distance_nuc) / expected_distance

    features += (index_mean_dist_nuc, index_median_dist_nuc)

    return features


def features_in_out_nucleus(rna_coord, rna_coord_out_nuc, check_input=True):
    """Compute nucleus related features.

    Parameters
    ----------
    rna_coord : np.ndarray, np.int64
        Coordinates of the detected RNAs with zyx or yx coordinates in the
        first 3 or 2 columns.
    rna_coord_out_nuc : np.ndarray, np.int64
        Coordinates of the detected RNAs with zyx or yx coordinates in the
        first 3 or 2 columns. Spots detected inside the nucleus are removed.
    check_input : bool
        Check input validity.

    Returns
    -------
    proportion_rna_in_nuc : float
        Proportion of RNAs detected inside the nucleus.
    nb_rna_out_nuc : float
        Number of RNAs detected outside the nucleus.
    nb_rna_in_nuc : float
        Number of RNAs detected inside the nucleus.

    """
    # check parameters
    stack.check_parameter(check_input=bool)
    if check_input:
        stack.check_array(rna_coord, ndim=2, dtype=np.int64)
        stack.check_array(rna_coord_out_nuc, ndim=2, dtype=np.int64)

    # count total number of rna
    nb_rna = float(len(rna_coord))

    # case where no rna are detected
    if nb_rna == 0:
        features = (0., 0., 0.)
        return features

    # number of rna outside and inside nucleus
    nb_rna_out_nuc = float(len(rna_coord_out_nuc))
    nb_rna_in_nuc = nb_rna - nb_rna_out_nuc

    # compute the proportion of rna in the nucleus
    proportion_rna_in_nuc = nb_rna_in_nuc / nb_rna

    features = (proportion_rna_in_nuc, nb_rna_out_nuc, nb_rna_in_nuc)

    return features


def features_protrusion(rna_coord, cell_mask, nuc_mask, ndim, voxel_size_yx,
                        check_input=True):
    """Compute protrusion related features.

    Parameters
    ----------
    rna_coord : np.ndarray, np.int64
        Coordinates of the detected RNAs with zyx or yx coordinates in the
        first 3 or 2 columns.
    cell_mask : np.ndarray, bool
        Surface of the cell with shape (y, x).
    nuc_mask : np.ndarray, bool
        Surface of the nucleus with shape (y, x).
    ndim : int
        Number of spatial dimensions to consider.
    voxel_size_yx : int or float
        Size of a voxel on the yx plan, in nanometer.
    check_input : bool
        Check input validity.

    Returns
    -------
    index_rna_protrusion : float
        Number of RNAs detected in a protrusion and normalized by the expected
        number of RNAs under random distribution.
    proportion_rna_protrusion : float
        Proportion of RNAs detected in a protrusion.
    protrusion_area : float
        Protrusion area (in pixels).

    """
    # TODO fin a better feature for the protrusion (idea: dilate region from
    #  centroid and stop when a majority of new pixels do not belong to the
    #  cell).

    # check parameters
    stack.check_parameter(check_input=bool)
    if check_input:
        stack.check_parameter(ndim=int,
                              voxel_size_yx=(int, float))
        if ndim not in [2, 3]:
            raise ValueError("'ndim' should be 2 or 3, not {0}.".format(ndim))
        stack.check_array(rna_coord, ndim=2, dtype=np.int64)
        stack.check_array(cell_mask, ndim=2, dtype=bool)
        stack.check_array(nuc_mask, ndim=2, dtype=bool)

    # get number of rna and cell area
    nb_rna = len(rna_coord)
    cell_area = cell_mask.sum()

    # apply opening operator (3000 nanometers) and count the loss of RNAs
    size = int(3000 / voxel_size_yx)
    s = disk(size, dtype=bool)
    mask_cell_opened = binary_opening(cell_mask, selem=s)
    mask_cell_opened[nuc_mask] = True
    protrusion_area = cell_area - mask_cell_opened.sum()

    # case where we do not detect any
    if nb_rna == 0:
        features = (1., 0., protrusion_area)
        return features

    if protrusion_area > 0:
        expected_rna_protrusion = nb_rna * protrusion_area / cell_area
        mask_rna = mask_cell_opened[rna_coord[:, ndim - 2],
                                    rna_coord[:, ndim - 1]]
        rna_after_opening = rna_coord[mask_rna]
        nb_rna_protrusion = nb_rna - len(rna_after_opening)
        index_rna_protrusion = nb_rna_protrusion / expected_rna_protrusion
        proportion_rna_protrusion = nb_rna_protrusion / nb_rna

        features = (index_rna_protrusion,
                    proportion_rna_protrusion,
                    protrusion_area)
    else:
        features = (1., 0., 0.)

    return features


def features_dispersion(smfish, rna_coord, centroid_rna, cell_mask,
                        centroid_cell, centroid_nuc, ndim, check_input=True):
    """Compute RNA Distribution Index features (RDI) described in:

    RDI Calculator: An analysis Tool to assess RNA distributions in cells,
    Stueland M., Wang T., Park H. Y., Mili, S., 2019.

    Parameters
    ----------
    smfish : np.ndarray, np.uint
        Image of RNAs, with shape (y, x).
    rna_coord : np.ndarray, np.int64
        Coordinates of the detected RNAs with zyx or yx coordinates in the
        first 3 or 2 columns.
    centroid_rna : np.ndarray, np.int64
        Coordinates of the rna centroid with shape (2,) or (3,).
    cell_mask : np.ndarray, bool
        Surface of the cell with shape (y, x).
    centroid_cell : np.ndarray, np.int64
        Coordinates of the cell centroid with shape (2,).
    centroid_nuc : np.ndarray, np.int64
        Coordinates of the nucleus centroid with shape (2,).
    ndim : int
        Number of spatial dimensions to consider.
    check_input : bool
        Check input validity.

    Returns
    -------
    index_polarization : float
        Polarization index (PI).
    index_dispersion : float
        Dispersion index (DI).
    index_peripheral_distribution : float
        Peripheral distribution index (PDI).

    """
    # check parameters
    stack.check_parameter(check_input=bool)
    if check_input:
        stack.check_parameter(ndim=int)
        if ndim not in [2, 3]:
            raise ValueError("'ndim' should be 2 or 3, not {0}.".format(ndim))
        stack.check_array(smfish, ndim=2, dtype=[np.uint8, np.uint16])
        stack.check_array(rna_coord, ndim=2, dtype=np.int64)
        stack.check_array(centroid_rna, ndim=1, dtype=np.int64)
        stack.check_array(cell_mask, ndim=2, dtype=bool)
        stack.check_array(centroid_cell, ndim=1, dtype=np.int64)
        stack.check_array(centroid_nuc, ndim=1, dtype=np.int64)

    # case where no mRNAs are detected
    if len(rna_coord) == 0:
        features = (0., 1., 1.)
        return features

    # initialization
    if ndim == 3:
        centroid_rna_2d = centroid_rna[1:]
    else:
        centroid_rna_2d = centroid_rna.copy()

    # get coordinates of each pixel of the cell
    cell_coord = np.nonzero(cell_mask)
    cell_coord = np.column_stack(cell_coord)

    # get coordinates of each rna pixel in the cell from a 2-d binary mask
    rna_mask = np.zeros_like(cell_mask)
    rna_mask[rna_coord[:, ndim - 2], rna_coord[:, ndim - 1]] = True
    rna_coord_ = np.nonzero(rna_mask)
    rna_coord_ = np.column_stack(rna_coord_)

    # get intensity value of every rna and cell pixels
    rna_value = smfish[rna_mask]
    total_intensity_rna = rna_value.sum()
    cell_value = smfish[cell_mask]
    total_intensity_cell = cell_value.sum()

    # compute polarization index from cell centroid
    centroid_distance = np.linalg.norm(centroid_rna_2d - centroid_cell)
    gyration_radius = _rmsd(cell_coord, centroid_cell)
    index_polarization = centroid_distance / gyration_radius

    features = (index_polarization,)

    # compute dispersion index
    r = np.linalg.norm(rna_coord_ - centroid_rna_2d, axis=1) ** 2
    a = np.sum((r * rna_value) / total_intensity_rna)
    r = np.linalg.norm(cell_coord - centroid_rna_2d, axis=1) ** 2
    b = np.sum((r * cell_value) / total_intensity_cell)
    index_dispersion = a / b

    features += (index_dispersion,)

    # compute peripheral distribution index
    r = np.linalg.norm(rna_coord_ - centroid_nuc, axis=1) ** 2
    a = np.sum((r * rna_value) / total_intensity_rna)
    r = np.linalg.norm(cell_coord - centroid_nuc, axis=1) ** 2
    b = np.sum((r * cell_value) / total_intensity_cell)
    index_peripheral_distribution = a / b

    features += (index_peripheral_distribution,)

    return features


def _rmsd(coord, reference_coord):
    """Compute the root-mean-squared distance between coordinates and a
    reference coordinate.

    Parameters
    ----------
    coord : np.ndarray, np.int64
        Coordinates with shape (nb_points, 2).
    reference_coord : np.ndarray, np.int64
        Reference coordinate to compute the distance from, with shape (2,).

    Returns
    -------
    rmsd : float
        Root-mean-squared distance.

    """
    # compute RMSD between 'coord' and 'reference_coord'
    n = len(coord)
    diff = coord - reference_coord
    rmsd = float(np.sqrt((diff ** 2).sum() / n))

    return rmsd


def features_topography(rna_coord, cell_mask, nuc_mask, cell_mask_out_nuc,
                        ndim, voxel_size_yx,
                        check_input=True):
    """Compute topographic features.

    Parameters
    ----------
    rna_coord : np.ndarray, np.int64
        Coordinates of the detected RNAs with zyx or yx coordinates in the
        first 3 or 2 columns.
    cell_mask : np.ndarray, bool
        Surface of the cell with shape (y, x).
    nuc_mask : np.ndarray, bool
        Surface of the nucleus with shape (y, x).
    cell_mask_out_nuc : np.ndarray, bool
        Surface of the cell (outside the nucleus) with shape (y, x).
    ndim : int
        Number of spatial dimensions to consider.
    voxel_size_yx : int or float
        Size of a voxel on the yx plan, in nanometer.
    check_input : bool
        Check input validity.

    Returns
    -------
    index_rna_nuc_marge : float
        Number of RNAs detected in a specific region around nucleus and
        normalized by the expected number of RNAs under random distribution.
        Six regions are targeted (less than 500nm, 500-1000nm, 1000-1500nm,
        1500-2000nm, 2000-2500nm and 2500-3000nm from the nucleus boundary).
    proportion_rna_nuc_marge : float
        Proportion of RNAs detected in a specific region around nucleus. Six
        regions are targeted (less than 500nm, 500-1000nm, 1000-1500nm,
        1500-2000nm, 2000-2500nm and 2500-3000nm from the nucleus boundary).
    index_rna_cell_marge : float
        Number of RNAs detected in a specific region around cell membrane and
        normalized by the expected number of RNAs under random distribution.
        Six regions are targeted (0-500nm, 500-1000nm, 1000-1500nm,
        1500-2000nm, 2000-2500nm and 2500-3000nm from the cell membrane).
    proportion_rna_cell_marge : float
        Proportion of RNAs detected in a specific region around cell membrane.
        Six regions are targeted (0-500nm, 500-1000nm, 1000-1500nm,
        1500-2000nm, 2000-2500nm and 2500-3000nm from the cell membrane).

    """
    # check parameters
    stack.check_parameter(check_input=bool)
    if check_input:
        stack.check_parameter(ndim=int,
                              voxel_size_yx=(int, float))
        if ndim not in [2, 3]:
            raise ValueError("'ndim' should be 2 or 3, not {0}.".format(ndim))
        stack.check_array(rna_coord, ndim=2, dtype=np.int64)
        stack.check_array(cell_mask, ndim=2, dtype=bool)
        stack.check_array(nuc_mask, ndim=2, dtype=bool)
        stack.check_array(cell_mask_out_nuc, ndim=2, dtype=bool)

    # initialization
    cell_area = cell_mask.sum()
    nb_rna = len(rna_coord)
    marge = int(500 / voxel_size_yx)
    if marge < 1:
        warnings.warn(UserWarning, "'voxel_size_yx' is greater than 500 "
                                   "nanometers ({0}). Topographic features "
                                   "use contour lines every 500 nanometers so "
                                   "they can't be computed here."
                      .format(voxel_size_yx))
        features = (0., 0.)
        features += (0., 0.) * 5
        features += (0., 0.) * 6
        return features

    # case where no mRNAs are detected
    if nb_rna == 0:
        features = (0., 0.)
        features += (0., 0.) * 5
        features += (0., 0.) * 6
        return features

    # build a distance map from nucleus boundary and from cell membrane
    distance_map_nuc_out = ndi.distance_transform_edt(~nuc_mask)
    distance_map_nuc_in = ndi.distance_transform_edt(~cell_mask_out_nuc)
    distance_map_nuc = distance_map_nuc_out + distance_map_nuc_in
    distance_map_nuc[~cell_mask] = 0.
    distance_map_cell = ndi.distance_transform_edt(cell_mask)

    # count mRNAs along nucleus edge (less than 500nm to the nucleus boundary)
    mask_nuc_marge = distance_map_nuc < marge
    mask_nuc_marge[~cell_mask] = False
    marge_area = max(mask_nuc_marge.sum(), 1)
    expected_rna_nuc_marge = nb_rna * marge_area / cell_area
    mask_rna = mask_nuc_marge[rna_coord[:, ndim-2], rna_coord[:, ndim-1]]
    nb_rna_nuc_marge = len(rna_coord[mask_rna])
    index_rna_nuc_marge = nb_rna_nuc_marge / expected_rna_nuc_marge
    proportion_rna_nuc_marge = nb_rna_nuc_marge / nb_rna

    features = (index_rna_nuc_marge, proportion_rna_nuc_marge)

    # count mRNAs in specific regions around nucleus (500-1000nm, 1000-1500nm,
    # 1500-2000nm, 2000-2500nm, 2500-3000nm)
    mask_cumulated_marge = mask_nuc_marge.copy()
    for i in range(2, 7):
        mask_nuc_marge = distance_map_nuc < i * marge
        mask_nuc_marge[~cell_mask] = False
        mask_nuc_marge[nuc_mask] = False
        mask_nuc_marge[mask_cumulated_marge] = False
        mask_cumulated_marge |= mask_nuc_marge
        marge_area = max(mask_nuc_marge.sum(), 1)
        expected_rna_nuc_marge = nb_rna * marge_area / cell_area
        mask_rna = mask_nuc_marge[rna_coord[:, ndim-2], rna_coord[:, ndim-1]]
        nb_rna_nuc_marge = len(rna_coord[mask_rna])
        index_rna_nuc_marge = nb_rna_nuc_marge / expected_rna_nuc_marge
        proportion_rna_nuc_marge = nb_rna_nuc_marge / nb_rna

        features += (index_rna_nuc_marge, proportion_rna_nuc_marge)

    # count mRNAs in specific regions around cell membrane (0-500nm,
    # 500-1000nm, 1000-1500nm, 1500-2000nm, 2000-2500nm, 2500-3000nm)
    mask_cumulated_marge = np.zeros_like(cell_mask)
    for i in range(1, 7):
        mask_cell_marge = distance_map_cell < i * marge
        mask_cell_marge[~cell_mask] = False
        mask_cell_marge[nuc_mask] = False
        mask_cell_marge[mask_cumulated_marge] = False
        mask_cumulated_marge |= mask_cell_marge
        marge_area = max(mask_cell_marge.sum(), 1)
        expected_rna_cell_marge = nb_rna * marge_area / cell_area
        mask_rna = mask_cell_marge[rna_coord[:, ndim-2], rna_coord[:, ndim-1]]
        nb_rna_cell_marge = len(rna_coord[mask_rna])
        index_rna_cell_marge = nb_rna_cell_marge / expected_rna_cell_marge
        proportion_rna_cell_marge = nb_rna_cell_marge / nb_rna

        features += (index_rna_cell_marge, proportion_rna_cell_marge)

    return features


def features_foci(rna_coord, foci_coord, ndim, check_input=True):
    """Compute foci related features.

    Parameters
    ----------
    rna_coord : np.ndarray, np.int64
        Coordinates of the detected RNAs with zyx or yx coordinates in the
        first 3 or 2 columns.
    foci_coord : np.ndarray, np.int64
        Array with shape (nb_foci, 5) or (nb_foci, 4). One coordinate per
        dimension for the foci centroid (zyx or yx coordinates), the number of
        spots detected in the foci and its index.
    ndim : int
        Number of spatial dimensions to consider.
    check_input : bool
        Check input validity.

    Returns
    -------
    proportion_rna_in_foci : float
        Proportion of RNAs detected in a foci.

    """
    # check parameters
    stack.check_parameter(check_input=bool)
    if check_input:
        stack.check_parameter(ndim=int)
        if ndim not in [2, 3]:
            raise ValueError("'ndim' should be 2 or 3, not {0}.".format(ndim))
        stack.check_array(rna_coord, ndim=2, dtype=np.int64)
        stack.check_array(foci_coord, ndim=2, dtype=np.int64)

    if len(rna_coord) == 0 or len(foci_coord) == 0:
        features = (0.,)
        return features

    # compute proportion RNAs in foci
    nb_rna = len(rna_coord)
    nb_rna_in_foci = foci_coord[:, ndim].sum()
    proportion_rna_in_foci = nb_rna_in_foci / nb_rna

    features = (proportion_rna_in_foci,)

    return features


def features_area(cell_mask, nuc_mask, cell_mask_out_nuc, check_input=True):
    """Compute area related features.

    Parameters
    ----------
    cell_mask : np.ndarray, bool
        Surface of the cell with shape (y, x).
    nuc_mask : np.ndarray, bool
        Surface of the nucleus with shape (y, x).
    cell_mask_out_nuc : np.ndarray, bool
        Surface of the cell (outside the nucleus) with shape (y, x).
    check_input : bool
        Check input validity.

    Returns
    -------
    nuc_relative_area : float
        Proportion of nucleus area in the cell.
    cell_area : float
        Cell area (in pixels).
    nuc_area : float
        Nucleus area (in pixels).
    cell_area_out_nuc : float
        Cell area outside the nucleus (in pixels).

    """
    # check parameters
    stack.check_parameter(check_input=bool)
    if check_input:
        stack.check_array(cell_mask, ndim=2, dtype=bool)
        stack.check_array(nuc_mask, ndim=2, dtype=bool)
        stack.check_array(cell_mask_out_nuc, ndim=2, dtype=bool)

    # get area of the cell and the nucleus
    cell_area = float(cell_mask.sum())
    nuc_area = float(nuc_mask.sum())

    # compute relative area of the nucleus
    nuc_relative_area = nuc_area / cell_area

    # compute area of the cell outside nucleus
    cell_area_out_nuc = float(cell_mask_out_nuc.sum())

    # return features
    features = (nuc_relative_area, cell_area, nuc_area, cell_area_out_nuc)

    return features


def features_centrosome(smfish, rna_coord, distance_centrosome, cell_mask,
                        ndim, voxel_size_yx, check_input=True):
    """Compute centrosome related features (in 2 dimensions).

    Parameters
    ----------
    smfish : np.ndarray, np.uint
        Image of RNAs, with shape (y, x).
    rna_coord : np.ndarray, np.int64
        Coordinates of the detected RNAs with zyx or yx coordinates in the
        first 3 or 2 columns.
    distance_centrosome : np.ndarray, np.float32
        Distance map from the centrosome with shape (y, x), in pixels.
    cell_mask : np.ndarray, bool
        Surface of the cell with shape (y, x).
    ndim : int
        Number of spatial dimensions to consider.
    voxel_size_yx : int or float
        Size of a voxel on the yx plan, in nanometer.
    check_input : bool
        Check input validity.

    Returns
    -------
    index_mean_dist_cent : float
        Normalized mean distance of RNAs to the closest centrosome.
    index_median_dist_cent : float
        Normalized median distance of RNAs to the closest centrosome.
    index_rna_centrosome : float
        Number of RNAs within a 2000nm radius from a centrosome, normalized by
        the expected number of RNAs under random distribution.
    proportion_rna_centrosome : float
        Proportion of RNAs within a 2000nm radius from a centrosome.
    index_centrosome_dispersion : float
        Centrosomal dispersion index. It quantify the dispersion of RNAs
        around centrosomes. The lower, the closer the RNAs are.

    """
    # check parameters
    stack.check_parameter(check_input=bool)
    if check_input:
        stack.check_parameter(ndim=int,
                              voxel_size_yx=(int, float))
        if ndim not in [2, 3]:
            raise ValueError("'ndim' should be 2 or 3, not {0}.".format(ndim))
        stack.check_array(smfish, ndim=2, dtype=[np.uint8, np.uint16])
        stack.check_array(rna_coord, ndim=2, dtype=np.int64)
        stack.check_array(distance_centrosome, ndim=2,
                          dtype=[np.float16, np.float32, np.float64])
        stack.check_array(cell_mask, ndim=2, dtype=bool)

    # case where no mRNAs are detected
    if len(rna_coord) == 0:
        features = (1., 1., 1., 0., 1.)
        return features

    # initialization
    nb_rna = len(rna_coord)
    cell_area = cell_mask.sum()

    # compute mean and median distances from the centrosomes
    rna_distance_cent = distance_centrosome[rna_coord[:, ndim - 2],
                                            rna_coord[:, ndim - 1]]
    expected_distance = np.mean(distance_centrosome[cell_mask])
    index_mean_dist_cent = np.mean(rna_distance_cent) / expected_distance
    expected_distance = np.median(distance_centrosome[cell_mask])
    index_median_dist_cent = np.median(rna_distance_cent) / expected_distance

    features = (index_mean_dist_cent, index_median_dist_cent)

    # compute proportion of mRNAs next to the centrosomes (<2000nm)
    radius = int(2000 / voxel_size_yx)
    if radius < 1:
        warnings.warn(UserWarning, "'voxel_size_yx' is greater than 2000 "
                                   "nanometers ({0}). Centrosome neighborhood "
                                   "is defined with a radius of 2000 "
                                   "nanometers so it can't be computed here."
                      .format(voxel_size_yx))
        features += (1, 0)
    else:
        mask_centrosome = distance_centrosome < radius
        mask_centrosome[~cell_mask] = False
        centrosome_area = max(mask_centrosome.sum(), 1)
        expected_nb_rna = nb_rna * centrosome_area / cell_area
        mask_rna = mask_centrosome[rna_coord[:, ndim - 2],
                                   rna_coord[:, ndim - 1]]
        nb_rna_centrosome = len(rna_coord[mask_rna])
        index_rna_centrosome = nb_rna_centrosome / expected_nb_rna
        proportion_rna_centrosome = nb_rna_centrosome / len(rna_coord)

        features += (index_rna_centrosome, proportion_rna_centrosome)

    # get coordinates of each pixel of the cell
    cell_coord = np.nonzero(cell_mask)
    cell_coord = np.column_stack(cell_coord)

    # get intensity value of every rna and cell pixels
    rna_value = smfish[rna_coord[:, ndim - 2], rna_coord[:, ndim - 1]]
    total_intensity_rna = rna_value.sum()
    cell_value = smfish[cell_coord[:, 0], cell_coord[:, 1]]
    total_intensity_cell = cell_value.sum()

    # compute attraction index
    r = distance_centrosome[rna_coord[:, ndim - 2],
                            rna_coord[:, ndim - 1]] ** 2
    a = np.sum((r * rna_value) / total_intensity_rna)
    r = distance_centrosome[cell_coord[:, 0], cell_coord[:, 1]] ** 2
    b = np.sum((r * cell_value) / total_intensity_cell)
    index_centrosome_dispersion = a / b

    features += (index_centrosome_dispersion,)

    return features



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for bigfish.stack.projection module.
"""

import pytest

import numpy as np
import bigfish.stack as stack

from bigfish.stack.projection import _one_hot_3d

from numpy.testing import assert_array_equal


# toy images
x = np.array(
    [[[1, 0, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 1, 1, 0],
      [0, 0, 0, 0, 0]],

     [[2, 0, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 1, 1, 0],
      [0, 0, 0, 0, 0]],

     [[3, 0, 0, 0, 0],
      [0, 3, 0, 0, 3],
      [0, 3, 0, 0, 0],
      [0, 3, 3, 3, 0],
      [0, 0, 0, 0, 0]]],
    dtype=np.uint8)

x_out_focus = np.array(
    [[[1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1]],

     [[1, 0, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 1, 1, 0],
      [0, 0, 0, 0, 0]],

     [[1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1]],

     [[2, 0, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 1, 1, 0],
      [0, 0, 0, 0, 0]],

     [[3, 0, 0, 0, 0],
      [0, 3, 0, 0, 3],
      [0, 3, 0, 0, 0],
      [0, 3, 3, 3, 0],
      [0, 0, 0, 0, 0]],

     [[1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1]]],
    dtype=np.uint8)


# ### test 2-d projection ###

def test_maximum_projection():
    # uint8
    y = stack.maximum_projection(x)
    expected_y = np.array([[3, 0, 0, 0, 0],
                           [0, 3, 0, 0, 3],
                           [0, 3, 0, 0, 0],
                           [0, 3, 3, 3, 0],
                           [0, 0, 0, 0, 0]], dtype=np.uint8)
    assert_array_equal(y, expected_y)
    assert y.dtype == np.uint8

    # uint16
    y = stack.maximum_projection(x.astype(np.uint16))
    expected_y = expected_y.astype(np.uint16)
    assert_array_equal(y, expected_y)
    assert y.dtype == np.uint16


def test_mean_projection():
    # uint8
    y = stack.mean_projection(x)
    expected_y = np.array([[2, 0, 0, 0, 0],
                           [0, 1, 0, 0, 1],
                           [0, 1, 0, 0, 0],
                           [0, 1, 1, 1, 0],
                           [0, 0, 0, 0, 0]], dtype=np.uint8)
    assert_array_equal(y, expected_y)
    assert y.dtype == np.uint8

    # uint16
    y = stack.mean_projection(x.astype(np.uint16))
    expected_y = expected_y.astype(np.uint16)
    assert_array_equal(y, expected_y)
    assert y.dtype == np.uint16


def test_median_projection():
    # uint8
    y = stack.median_projection(x)
    expected_y = np.array([[2, 0, 0, 0, 0],
                           [0, 1, 0, 0, 0],
                           [0, 1, 0, 0, 0],
                           [0, 1, 1, 1, 0],
                           [0, 0, 0, 0, 0]], dtype=np.uint8)
    assert_array_equal(y, expected_y)
    assert y.dtype == np.uint8

    # uint16
    y = stack.median_projection(x.astype(np.uint16))
    expected_y = expected_y.astype(np.uint16)
    assert_array_equal(y, expected_y)
    assert y.dtype == np.uint16


# ### test focus selection ###

@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64])
def test_one_hot_3d(dtype):
    # integer
    indices = np.array(
        [[1, 0, 0, 0, 0],
         [0, 1, 0, 0, 0],
         [2, 1, 0, 2, 2],
         [0, 1, 1, 1, 2],
         [0, 0, 0, 1, 0]], dtype=dtype)
    one_hot = _one_hot_3d(indices=indices, depth=4)
    expected_one_hot = np.array(
        [[[0, 1, 1, 1, 1],
          [1, 0, 1, 1, 1],
          [0, 0, 1, 0, 0],
          [1, 0, 0, 0, 0],
          [1, 1, 1, 0, 1]],

         [[1, 0, 0, 0, 0],
          [0, 1, 0, 0, 0],
          [0, 1, 0, 0, 0],
          [0, 1, 1, 1, 0],
          [0, 0, 0, 1, 0]],

         [[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [1, 0, 0, 1, 1],
          [0, 0, 0, 0, 1],
          [0, 0, 0, 0, 0]],

         [[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0]]],
        dtype=np.uint8)
    assert_array_equal(one_hot, expected_one_hot)
    assert one_hot.dtype == indices.dtype

    # boolean
    one_hot = _one_hot_3d(indices=indices, depth=4, return_boolean=True)
    expected_one_hot = expected_one_hot.astype(bool)
    assert_array_equal(one_hot, expected_one_hot)
    assert one_hot.dtype == bool

# TODO remove test bigfish.stack.test_one_hot_3d
# TODO test bigfish.stack.focus_projection
# TODO test bigfish.stack.in_focus_selection
# TODO test bigfish.stack.get_in_focus_indices



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for bigfish.detection.spot_detection module.
"""

# TODO test bigfish.detection.detect_spots
# TODO test bigfish.detection.local_maximum_detection
# TODO test bigfish.detection.spots_thresholding
# TODO test bigfish.detection.automated_threshold_setting



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Functions to detect dense and bright regions (with potential clustered spots),
then use gaussian functions to correct a misdetection in these regions.
"""

import warnings

import numpy as np

import bigfish.stack as stack
from .spot_modeling import build_reference_spot, modelize_spot, precompute_erf
from .spot_modeling import gaussian_2d, _initialize_grid_2d
from .spot_modeling import gaussian_3d, _initialize_grid_3d

from skimage.measure import regionprops
from skimage.measure import label


# ### Main function ###

def decompose_dense(image, spots, voxel_size_z=None, voxel_size_yx=100,
                    psf_z=None, psf_yx=200, alpha=0.5, beta=1, gamma=5):
    """Detect dense and bright regions with potential clustered spots and
    simulate a more realistic number of spots in these regions.

    #. We estimate image background with a large gaussian filter. We then
       remove the background from the original image to denoise it.
    #. We build a reference spot by aggregating predetected spots.
    #. We fit gaussian parameters on the reference spots.
    #. We detect dense regions to decompose.
    #. We simulate as many gaussians as possible in the candidate regions.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x) or (y, x).
    spots : np.ndarray, np.int64
        Coordinate of the spots with shape (nb_spots, 3) or (nb_spots, 2)
        for 3-d or 2-d images respectively.
    voxel_size_z : int or float or None
        Height of a voxel, along the z axis, in nanometer. If None, image is
        considered in 2-d.
    voxel_size_yx : int or float
        Size of a voxel on the yx plan, in nanometer.
    psf_z : int or float or None
        Theoretical size of the PSF emitted by a spot in the z plan, in
        nanometer. If None, image is considered in 2-d.
    psf_yx : int or float
        Theoretical size of the PSF emitted by a spot in the yx plan, in
        nanometer.
    alpha : int or float
        Intensity percentile used to compute the reference spot, between 0
        and 1. The higher, the brighter are the spots simulated in the dense
        regions. Consequently, a high intensity score reduces the number of
        spots added. Default is 0.5, meaning the reference spot considered is
        the median spot.
    beta : int or float
        Multiplicative factor for the intensity threshold of a dense region.
        Default is 1. Threshold is computed with the formula:

        .. math::
            \\mbox{threshold} = \\beta * \\mbox{max(median spot)}

        With :math:`\\mbox{median spot}` the median value of all detected spot
        signals.
    gamma : int or float
        Multiplicative factor use to compute a gaussian scale:

        .. math::
            \\mbox{scale} = \\frac{\\gamma * \\mbox{PSF}}{\\mbox{voxel size}}

        We perform a large gaussian filter with such scale to estimate image
        background and remove it from original image. A large gamma increases
        the scale of the gaussian filter and smooth the estimated background.
        To decompose very large bright areas, a larger gamma should be set.
        If 0, image is not denoised.

    Returns
    -------
    spots : np.ndarray, np.int64
        Coordinate of the spots detected, with shape (nb_spots, 3) or
        (nb_spots, 2). One coordinate per dimension (zyx or yx coordinates).
    dense_regions : np.ndarray, np.int64
        Array with shape (nb_regions, 7) or (nb_regions, 6). One coordinate
        per dimension for the region centroid (zyx or yx coordinates), the
        number of RNAs detected in the region, the area of the region, its
        average intensity value and its index.
    reference_spot : np.ndarray
        Reference spot in 3-d or 2-d.

    """
    # check parameters
    stack.check_array(image,
                      ndim=[2, 3],
                      dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_array(spots, ndim=2, dtype=np.int64)
    stack.check_parameter(voxel_size_z=(int, float, type(None)),
                          voxel_size_yx=(int, float),
                          psf_z=(int, float, type(None)),
                          psf_yx=(int, float),
                          alpha=(int, float),
                          beta=(int, float),
                          gamma=(int, float))
    if alpha < 0 or alpha > 1:
        raise ValueError("'alpha' should be a value between 0 and 1, not {0}"
                         .format(alpha))
    if beta < 0:
        raise ValueError("'beta' should be a positive value, not {0}"
                         .format(beta))
    if gamma < 0:
        raise ValueError("'gamma' should be a positive value, not {0}"
                         .format(gamma))

    # check number of dimensions
    ndim = image.ndim
    if ndim == 3 and voxel_size_z is None:
        raise ValueError("Provided image has {0} dimensions but "
                         "'voxel_size_z' parameter is missing.".format(ndim))
    if ndim == 3 and psf_z is None:
        raise ValueError("Provided image has {0} dimensions but "
                         "'psf_z' parameter is missing.".format(ndim))
    if ndim != spots.shape[1]:
        raise ValueError("Provided image has {0} dimensions but spots are "
                         "detected in {1} dimensions."
                         .format(ndim, spots.shape[1]))
    if ndim == 2:
        voxel_size_z, psf_z = None, None

    # case where no spot were detected
    if spots.size == 0:
        dense_regions = np.array([], dtype=np.int64).reshape((0, ndim + 4))
        reference_spot = np.zeros((5,) * ndim, dtype=image.dtype)
        return spots, dense_regions, reference_spot

    # compute expected standard deviation of the spots
    sigma = stack.get_sigma(voxel_size_z, voxel_size_yx, psf_z, psf_yx)
    large_sigma = tuple([sigma_ * gamma for sigma_ in sigma])

    # denoise the image
    if gamma > 0:
        image_denoised = stack.remove_background_gaussian(
            image,
            sigma=large_sigma)
    else:
        image_denoised = image.copy()

    # build a reference median spot
    reference_spot = build_reference_spot(
        image_denoised,
        spots,
        voxel_size_z, voxel_size_yx, psf_z, psf_yx,
        alpha)

    # case with an empty frame as reference spot
    if reference_spot.sum() == 0:
        dense_regions = np.array([], dtype=np.int64).reshape((0, ndim + 4))
        return spots, dense_regions, reference_spot

    # fit a gaussian function on the reference spot to be able to simulate it
    parameters_fitted = modelize_spot(
        reference_spot, voxel_size_z, voxel_size_yx, psf_z, psf_yx)
    if ndim == 3:
        sigma_z, sigma_yx, amplitude, background = parameters_fitted
    else:
        sigma_z = None
        sigma_yx, amplitude, background = parameters_fitted

    # use connected components to detect dense and bright regions
    regions_to_decompose, spots_out_regions, region_size = get_dense_region(
        image_denoised,
        spots,
        voxel_size_z, voxel_size_yx, psf_z, psf_yx,
        beta)

    # case where no region where detected
    if regions_to_decompose.size == 0:
        dense_regions = np.array([], dtype=np.int64).reshape((0, ndim + 4))
        return spots, dense_regions, reference_spot

    # precompute gaussian function values
    max_grid = region_size + 1
    precomputed_gaussian = precompute_erf(
        voxel_size_z, voxel_size_yx, sigma_z, sigma_yx, max_grid=max_grid)

    # simulate gaussian mixtures in the dense regions
    spots_in_regions, dense_regions = simulate_gaussian_mixture(
        image=image_denoised,
        candidate_regions=regions_to_decompose,
        voxel_size_z=voxel_size_z,
        voxel_size_yx=voxel_size_yx,
        sigma_z=sigma_z,
        sigma_yx=sigma_yx,
        amplitude=amplitude,
        background=background,
        precomputed_gaussian=precomputed_gaussian)

    # normally the number of detected spots should increase
    if len(spots_out_regions) + len(spots_in_regions) < len(spots):
        warnings.warn("Problem occurs during the decomposition of dense "
                      "regions. Less spots are detected after the "
                      "decomposition than before.",
                      UserWarning)

    # merge outside and inside spots
    spots = np.concatenate((spots_out_regions, spots_in_regions[:, :ndim]),
                           axis=0)

    return spots, dense_regions, reference_spot


# ### Dense regions ###

def get_dense_region(image, spots, voxel_size_z=None, voxel_size_yx=100,
                     psf_z=None, psf_yx=200, beta=1):
    """Detect and filter dense and bright regions.

    A candidate region has at least 2 connected pixels above a specific
    threshold.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x) or (y, x).
    spots : np.ndarray, np.int64
        Coordinate of the spots with shape (nb_spots, 3) or (nb_spots, 2).
    voxel_size_z : int or float or None
        Height of a voxel, along the z axis, in nanometer. If None, we
        consider a 2-d image.
    voxel_size_yx : int or float
        Size of a voxel on the yx plan, in nanometer.
    psf_z : int or float or None
        Theoretical size of the PSF emitted by a spot in the z plan, in
        nanometer. If None, we consider a 2-d image.
    psf_yx : int or float
        Theoretical size of the PSF emitted by a spot in the yx plan, in
        nanometer.
    beta : int or float
        Multiplicative factor for the intensity threshold of a dense region.
        Default is 1. Threshold is computed with the formula:

        .. math::
            \\mbox{threshold} = \\beta * \\mbox{max(median spot)}

        With :math:`\\mbox{median spot}` the median value of all detected spot
        signals.

    Returns
    -------
    dense_regions : np.ndarray
        Array with selected ``skimage.measure._regionprops._RegionProperties``
        objects.
    spots_out_region : np.ndarray, np.int64
        Coordinate of the spots detected out of dense regions, with shape
        (nb_spots, 3) or (nb_spots, 2). One coordinate per dimension (zyx or
        yx coordinates).
    max_size : int
        Maximum size of the regions.

    """
    # check parameters
    stack.check_array(image,
                      ndim=[2, 3],
                      dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_array(spots, ndim=2, dtype=np.int64)
    stack.check_parameter(voxel_size_z=(int, float, type(None)),
                          voxel_size_yx=(int, float),
                          psf_z=(int, float, type(None)),
                          psf_yx=(int, float),
                          beta=(int, float))
    if beta < 0:
        raise ValueError("'beta' should be a positive value, not {0}"
                         .format(beta))

    # check number of dimensions
    ndim = image.ndim
    if ndim == 3 and voxel_size_z is None:
        raise ValueError("Provided image has {0} dimensions but "
                         "'voxel_size_z' parameter is missing.".format(ndim))
    if ndim == 3 and psf_z is None:
        raise ValueError("Provided image has {0} dimensions but "
                         "'psf_z' parameter is missing.".format(ndim))
    if ndim != spots.shape[1]:
        raise ValueError("Provided image has {0} dimensions but spots are "
                         "detected in {1} dimensions."
                         .format(ndim, spots.shape[1]))
    if ndim == 2:
        voxel_size_z, psf_z = None, None

    # estimate median spot value and a threshold to detect dense regions
    median_spot = build_reference_spot(
        image,
        spots,
        voxel_size_z, voxel_size_yx, psf_z, psf_yx,
        alpha=0.5)
    threshold = int(median_spot.max() * beta)

    # get connected regions
    connected_regions = _get_connected_region(image, threshold)

    # filter connected regions
    (dense_regions, spots_out_region, max_size) = _filter_connected_region(
        image, connected_regions, spots)

    return dense_regions, spots_out_region, max_size


def _get_connected_region(image, threshold):
    """Find connected regions above a fixed threshold.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x) or (y, x).
    threshold : int or float
        A threshold to detect peaks.

    Returns
    -------
    cc : np.ndarray, np.int64
        Image labelled with shape (z, y, x) or (y, x).

    """
    # compute binary mask of the filtered image
    mask = image > threshold

    # find connected components
    cc = label(mask)

    return cc


def _filter_connected_region(image, connected_component, spots):
    """Filter dense and bright regions (defined as connected component
    regions).

    A candidate region has at least 2 connected pixels above a specific
    threshold.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x) or (y, x).
    connected_component : np.ndarray, np.int64
        Image labelled with shape (z, y, x) or (y, x).
    spots : np.ndarray, np.int64
        Coordinate of the spots with shape (nb_spots, 3) or (nb_spots, 2).

    Returns
    -------
    regions_filtered : np.ndarray
        Array with filtered skimage.measure._regionprops._RegionProperties.
    spots_out_region : np.ndarray, np.int64
        Coordinate of the spots outside the regions with shape (nb_spots, 3)
        or (nb_spots, 2).
    max_region_size : int
        Maximum size of the regions.

    """
    # get properties of the different connected regions
    regions = regionprops(connected_component, intensity_image=image)

    # get different features of the regions
    area = []
    bbox = []
    for i, region in enumerate(regions):
        area.append(region.area)
        bbox.append(region.bbox)
    regions = np.array(regions)
    area = np.array(area)
    bbox = np.array(bbox)

    # keep regions with a minimum size
    big_area = area >= 2
    regions_filtered = regions[big_area]
    bbox_filtered = bbox[big_area]

    # case where no region big enough were detected
    if regions.size == 0:
        regions_filtered = np.array([])
        return regions_filtered, spots, 0

    spots_out_region, max_region_size = _filter_spot_out_candidate_regions(
        bbox_filtered, spots, nb_dim=image.ndim)

    return regions_filtered, spots_out_region, max_region_size


def _filter_spot_out_candidate_regions(candidate_bbox, spots, nb_dim):
    """Filter spots out of the dense regions.

    Parameters
    ----------
    candidate_bbox : List[Tuple]
        List of Tuples with the bounding box coordinates.
    spots : np.ndarray, np.int64
        Coordinate of the spots with shape (nb_spots, 3) or (nb_spots, 2).
    nb_dim : int
        Number of dimensions to consider (2 or 3).

    Returns
    -------
    spots_out_region : np.ndarray, np.int64
        Coordinate of the spots outside the regions with shape (nb_spots, 3)
        or (nb_spots, 2).
    max_region_size : int
        Maximum size of the regions.

    """
    # initialization
    mask_spots_out = np.ones(spots[:, 0].shape, dtype=bool)
    max_region_size = 0

    # get detected spots outside 3-d regions
    if nb_dim == 3:
        for box in candidate_bbox:
            (min_z, min_y, min_x, max_z, max_y, max_x) = box

            # get the size of the biggest region
            size_z = max_z - min_z
            size_y = max_y - min_y
            size_x = max_x - min_x
            max_region_size = max(max_region_size, size_z, size_y, size_x)

            # get coordinates of spots inside the region
            mask_spots_in = spots[:, 0] < max_z
            mask_spots_in = (mask_spots_in & (spots[:, 1] < max_y))
            mask_spots_in = (mask_spots_in & (spots[:, 2] < max_x))
            mask_spots_in = (mask_spots_in & (min_z <= spots[:, 0]))
            mask_spots_in = (mask_spots_in & (min_y <= spots[:, 1]))
            mask_spots_in = (mask_spots_in & (min_x <= spots[:, 2]))
            mask_spots_out = mask_spots_out & (~mask_spots_in)

    # get detected spots outside 2-d regions
    else:
        for box in candidate_bbox:
            (min_y, min_x, max_y, max_x) = box

            # get the size of the biggest region
            size_y = max_y - min_y
            size_x = max_x - min_x
            max_region_size = max(max_region_size, size_y, size_x)

            # get coordinates of spots inside the region
            mask_spots_in = spots[:, 0] < max_y
            mask_spots_in = (mask_spots_in & (spots[:, 1] < max_x))
            mask_spots_in = (mask_spots_in & (min_y <= spots[:, 0]))
            mask_spots_in = (mask_spots_in & (min_x <= spots[:, 1]))
            mask_spots_out = mask_spots_out & (~mask_spots_in)

    # keep apart spots inside a region
    spots_out_region = spots.copy()
    spots_out_region = spots_out_region[mask_spots_out]

    return spots_out_region, int(max_region_size)


# ### Gaussian simulation ###

def simulate_gaussian_mixture(image, candidate_regions, voxel_size_z=None,
                              voxel_size_yx=100, sigma_z=None, sigma_yx=200,
                              amplitude=100, background=0,
                              precomputed_gaussian=None):
    """Simulate as many gaussians as possible in the candidate dense regions in
    order to get a more realistic number of spots.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x) or (y, x).
    candidate_regions : np.ndarray
        Array with filtered skimage.measure._regionprops._RegionProperties.
    voxel_size_z : int or float or None
        Height of a voxel, along the z axis, in nanometer. If None, we consider
        a 2-d image.
    voxel_size_yx : int or float
        Size of a voxel on the yx plan, in nanometer.
    sigma_z : int or float or None
        Standard deviation of the gaussian along the z axis, in nanometer. If
        None, we consider a 2-d image.
    sigma_yx : int or float
        Standard deviation of the gaussian along the yx axis, in nanometer.
    amplitude : float
        Amplitude of the gaussian.
    background : float
        Background minimum value of the image.
    precomputed_gaussian : Tuple[np.ndarray]
        Tuple with one tables of precomputed values for the erf, with shape
        (nb_value, 2). One table per dimension.

    Returns
    -------
    spots_in_regions : np.ndarray, np.int64
        Coordinate of the spots detected inside dense regions, with shape
        (nb_spots, 4) or (nb_spots, 3). One coordinate per dimension (zyx
        or yx coordinates) plus the index of the region.
    regions : np.ndarray, np.int64
        Array with shape (nb_regions, 7) or (nb_regions, 6). One coordinate
        per dimension for the region centroid (zyx or yx coordinates), the
        number of RNAs detected in the region, the area of the region, its
        average intensity value and its index.

    """
    # check parameters
    stack.check_array(image,
                      ndim=[2, 3],
                      dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_parameter(candidate_regions=np.ndarray,
                          voxel_size_z=(int, float, type(None)),
                          voxel_size_yx=(int, float),
                          sigma_z=(int, float, type(None)),
                          sigma_yx=(int, float),
                          amplitude=float,
                          background=float)
    if background < 0:
        raise ValueError("Background value can't be negative: {0}"
                         .format(background))

    # check number of dimensions
    ndim = image.ndim
    if ndim == 3 and voxel_size_z is None:
        raise ValueError("Provided image has {0} dimensions but "
                         "'voxel_size_z' parameter is missing."
                         .format(ndim))
    if ndim == 3 and sigma_z is None:
        raise ValueError("Provided image has {0} dimensions but "
                         "'sigma_z' parameter is missing.".format(ndim))
    if ndim == 2:
        voxel_size_z, sigma_z = None, None

    # simulate gaussian mixtures in the candidate regions...
    spots_in_regions = []
    regions = []

    # ... for 3-d regions...
    if image.ndim == 3:

        for i_region, region in enumerate(candidate_regions):
            image_region, _, coord_gaussian = _gaussian_mixture_3d(
                image,
                region,
                voxel_size_z,
                voxel_size_yx,
                sigma_z,
                sigma_yx,
                amplitude,
                background,
                precomputed_gaussian)

            # get coordinates of spots and regions in the original image
            box = region.bbox
            (min_z, min_y, min_x, _, _, _) = box
            coord = np.array(coord_gaussian, dtype=np.float64)
            coord[:, 0] = (coord[:, 0] / voxel_size_z) + min_z
            coord[:, 1] = (coord[:, 1] / voxel_size_yx) + min_y
            coord[:, 2] = (coord[:, 2] / voxel_size_yx) + min_x
            spots_in_region = np.zeros((coord.shape[0], 4), dtype=np.int64)
            spots_in_region[:, :3] = coord
            spots_in_region[:, 3] = i_region
            spots_in_regions.append(spots_in_region)
            region_z, region_y, region_x = tuple(coord[0])
            nb_rna_region = coord.shape[0]
            region_area = region.area
            region_intensity = region.mean_intensity
            regions.append([region_z, region_y, region_x, nb_rna_region,
                            region_area, region_intensity, i_region])

    # ... or 2-d regions
    else:

        for i_region, region in enumerate(candidate_regions):
            image_region, _, coord_gaussian = _gaussian_mixture_2d(
                image,
                region,
                voxel_size_yx,
                sigma_yx,
                amplitude,
                background,
                precomputed_gaussian)

            # get coordinates of spots and regions in the original image
            box = region.bbox
            (min_y, min_x, _, _) = box
            coord = np.array(coord_gaussian, dtype=np.float64)
            coord[:, 0] = (coord[:, 0] / voxel_size_yx) + min_y
            coord[:, 1] = (coord[:, 1] / voxel_size_yx) + min_x
            spots_in_region = np.zeros((coord.shape[0], 3), dtype=np.int64)
            spots_in_region[:, :2] = coord
            spots_in_region[:, 2] = i_region
            spots_in_regions.append(spots_in_region)
            region_y, region_x = tuple(coord[0])
            nb_rna_region = coord.shape[0]
            region_area = region.area
            region_intensity = region.mean_intensity
            regions.append([region_y, region_x, nb_rna_region,
                            region_area, region_intensity, i_region])

    spots_in_regions = np.concatenate(spots_in_regions, axis=0)
    regions = np.array(regions, dtype=np.int64)

    return spots_in_regions, regions


def _gaussian_mixture_3d(image, region, voxel_size_z, voxel_size_yx, sigma_z,
                         sigma_yx, amplitude, background, precomputed_gaussian,
                         limit_gaussian=1000):
    """Fit as many 3-d gaussians as possible in a candidate region.

    Parameters
    ----------
    image : np.ndarray, np.uint
        A 3-d image with detected spot and shape (z, y, x).
    region : skimage.measure._regionprops._RegionProperties
        Properties of a candidate region.
    voxel_size_z : int or float
        Height of a voxel, along the z axis, in nanometer.
    voxel_size_yx : int or float
        Size of a voxel on the yx plan, in nanometer.
    sigma_z : int or float
        Standard deviation of the gaussian along the z axis, in pixel.
    sigma_yx : int or float
        Standard deviation of the gaussian along the yx axis, in pixel.
    amplitude : float
        Amplitude of the gaussian.
    background : float
        Background minimum value of the image.
    precomputed_gaussian : Tuple[np.ndarray]
        Tuple with one tables of precomputed values for the erf, with shape
        (nb_value, 2). One table per dimension.
    limit_gaussian : int
        Limit number of gaussian to fit into this region.

    Returns
    -------
    image_region : np.ndarray, np.uint
        A 3-d image with detected spots and shape (z, y, x).
    best_simulation : np.ndarray, np.uint
        A 3-d image with simulated spots and shape (z, y, x).
    positions_gaussian : List[List]
        List of positions (as a list [z, y, x]) for the different gaussian
        simulations used in the mixture.

    """
    # get an image of the region
    box = tuple(region.bbox)
    image_region = image[box[0]:box[3], box[1]:box[4], box[2]:box[5]]
    image_region_raw = np.reshape(image_region, image_region.size)
    image_region_raw = image_region_raw.astype(np.float64)

    # build a grid to represent this image
    grid = _initialize_grid_3d(image_region, voxel_size_z, voxel_size_yx)

    # add a gaussian for each local maximum while the RSS decreases
    simulation = np.zeros_like(image_region_raw)
    residual = image_region_raw - simulation
    ssr = np.sum(residual ** 2)
    diff_ssr = -1
    nb_gaussian = 0
    best_simulation = simulation.copy()
    positions_gaussian = []
    while diff_ssr < 0 or nb_gaussian == limit_gaussian:
        position_gaussian = np.argmax(residual)
        positions_gaussian.append(list(grid[:, position_gaussian]))
        simulation += gaussian_3d(grid=grid,
                                  mu_z=float(positions_gaussian[-1][0]),
                                  mu_y=float(positions_gaussian[-1][1]),
                                  mu_x=float(positions_gaussian[-1][2]),
                                  sigma_z=sigma_z,
                                  sigma_yx=sigma_yx,
                                  voxel_size_z=voxel_size_z,
                                  voxel_size_yx=voxel_size_yx,
                                  psf_amplitude=amplitude,
                                  psf_background=background,
                                  precomputed=precomputed_gaussian)
        residual = image_region_raw - simulation
        new_ssr = np.sum(residual ** 2)
        diff_ssr = new_ssr - ssr
        ssr = new_ssr
        nb_gaussian += 1
        background = 0

        if diff_ssr < 0:
            best_simulation = simulation.copy()

    if 1 < nb_gaussian < limit_gaussian:
        positions_gaussian.pop(-1)
    elif nb_gaussian == limit_gaussian:
        warnings.warn("Problem occurs during the decomposition of a dense "
                      "region. More than {0} spots seem to be necessary to "
                      "reproduce the candidate region and decomposition was "
                      "stopped early. Set a higher limit or check a potential "
                      "artifact in the image if you do not expect such a "
                      "large region to be decomposed.".format(limit_gaussian),
                      UserWarning)

    best_simulation = np.reshape(best_simulation, image_region.shape)
    max_value_dtype = np.iinfo(image_region.dtype).max
    best_simulation = np.clip(best_simulation, 0, max_value_dtype)
    best_simulation = best_simulation.astype(image_region.dtype)

    return image_region, best_simulation, positions_gaussian


def _gaussian_mixture_2d(image, region, voxel_size_yx, sigma_yx, amplitude,
                         background, precomputed_gaussian,
                         limit_gaussian=1000):
    """Fit as many 2-d gaussians as possible in a candidate region.

    Parameters
    ----------
    image : np.ndarray, np.uint
        A 2-d image with detected spot and shape (y, x).
    region : skimage.measure._regionprops._RegionProperties
        Properties of a candidate region.
    voxel_size_yx : int or float
        Size of a voxel on the yx plan, in nanometer.
    sigma_yx : int or float
        Standard deviation of the gaussian along the yx axis, in pixel.
    amplitude : float
        Amplitude of the gaussian.
    background : float
        Background minimum value of the image.
    precomputed_gaussian : Tuple[np.ndarray]
        Tuple with one tables of precomputed values for the erf, with shape
        (nb_value, 2). One table per dimension.
    limit_gaussian : int
        Limit number of gaussian to fit into this region.

    Returns
    -------
    image_region : np.ndarray, np.uint
        A 2-d image with detected spots and shape (y, x).
    best_simulation : np.ndarray, np.uint
        A 2-d image with simulated spots and shape (y, x).
    positions_gaussian : List[List]
        List of positions (as a list [y, x]) for the different gaussian
        simulations used in the mixture.

    """
    # get an image of the region
    box = tuple(region.bbox)
    image_region = image[box[0]:box[2], box[1]:box[3]]
    image_region_raw = np.reshape(image_region, image_region.size)
    image_region_raw = image_region_raw.astype(np.float64)

    # build a grid to represent this image
    grid = _initialize_grid_2d(image_region, voxel_size_yx)

    # add a gaussian for each local maximum while the RSS decreases
    simulation = np.zeros_like(image_region_raw)
    residual = image_region_raw - simulation
    ssr = np.sum(residual ** 2)
    diff_ssr = -1
    nb_gaussian = 0
    best_simulation = simulation.copy()
    positions_gaussian = []
    while diff_ssr < 0 or nb_gaussian == limit_gaussian:
        position_gaussian = np.argmax(residual)
        positions_gaussian.append(list(grid[:, position_gaussian]))
        simulation += gaussian_2d(grid=grid,
                                  mu_y=float(positions_gaussian[-1][0]),
                                  mu_x=float(positions_gaussian[-1][1]),
                                  sigma_yx=sigma_yx,
                                  voxel_size_yx=voxel_size_yx,
                                  psf_amplitude=amplitude,
                                  psf_background=background,
                                  precomputed=precomputed_gaussian)
        residual = image_region_raw - simulation
        new_ssr = np.sum(residual ** 2)
        diff_ssr = new_ssr - ssr
        ssr = new_ssr
        nb_gaussian += 1
        background = 0

        if diff_ssr < 0:
            best_simulation = simulation.copy()

    if 1 < nb_gaussian < limit_gaussian:
        positions_gaussian.pop(-1)
    elif nb_gaussian == limit_gaussian:
        warnings.warn("Problem occurs during the decomposition of a dense "
                      "region. More than {0} spots seem to be necessary to "
                      "reproduce the candidate region and decomposition was "
                      "stopped early. Set a higher limit or check a potential "
                      "artifact in the image if you do not expect such a "
                      "large region to be decomposed.".format(limit_gaussian),
                      UserWarning)

    best_simulation = np.reshape(best_simulation, image_region.shape)
    max_value_dtype = np.iinfo(image_region.dtype).max
    best_simulation = np.clip(best_simulation, 0, max_value_dtype)
    best_simulation = best_simulation.astype(image_region.dtype)

    return image_region, best_simulation, positions_gaussian



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for bigfish.segmentation.nuc_segmentation module.
"""


# TODO add test for bigfish.segmentation.remove_segmented_nuc



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for bigfish.stack.filter module.
"""

import pytest

import numpy as np
import bigfish.stack as stack

from bigfish.stack.filter import _define_kernel

from numpy.testing import assert_array_equal
from numpy.testing import assert_allclose


# toy images
x = np.array(
    [[3, 2, 0, 0, 0],
     [2, 1, 0, 0, 0],
     [0, 1, 0, 0, 0],
     [0, 2, 1, 5, 0],
     [0, 0, 0, 0, 0]],
    dtype=np.uint8)
y = np.array(
    [[0, 0, 62, 164, 55],
     [0, 0, 120, 235, 181],
     [0, 0, 73, 205, 0],
     [0, 131, 0, 0, 0],
     [0, 0, 0, 0, 0]],
    dtype=np.uint8)


@pytest.mark.parametrize("shape, size", [
    ("diamond", 3), ("disk", 3), ("rectangle", (2, 3)), ("square", 3),
    ("blabla", 3)])
@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float16, np.float32, np.float64, bool])
def test_kernel(shape, size, dtype):
    # non valid case
    if shape not in ["diamond", "disk", "rectangle", "square"]:
        with pytest.raises(ValueError):
            _define_kernel(shape, size, dtype)

    # valid cases
    else:
        kernel = _define_kernel(shape, size, dtype)
        if shape == "diamond":
            expected_kernel = np.array(
                [[0, 0, 0, 1, 0, 0, 0],
                 [0, 0, 1, 1, 1, 0, 0],
                 [0, 1, 1, 1, 1, 1, 0],
                 [1, 1, 1, 1, 1, 1, 1],
                 [0, 1, 1, 1, 1, 1, 0],
                 [0, 0, 1, 1, 1, 0, 0],
                 [0, 0, 0, 1, 0, 0, 0]],
                dtype=dtype)
        elif shape == "disk":
            expected_kernel = np.array(
                [[0, 0, 0, 1, 0, 0, 0],
                 [0, 1, 1, 1, 1, 1, 0],
                 [0, 1, 1, 1, 1, 1, 0],
                 [1, 1, 1, 1, 1, 1, 1],
                 [0, 1, 1, 1, 1, 1, 0],
                 [0, 1, 1, 1, 1, 1, 0],
                 [0, 0, 0, 1, 0, 0, 0]],
                dtype=dtype)
        elif shape == "rectangle":
            expected_kernel = np.array(
                [[1, 1, 1],
                 [1, 1, 1]],
                dtype=dtype)
        else:
            expected_kernel = np.array(
                [[1, 1, 1],
                 [1, 1, 1],
                 [1, 1, 1]],
                dtype=dtype)
        assert_array_equal(kernel, expected_kernel)
        assert kernel.dtype == dtype


def test_mean_filter():
    # np.uint8
    filtered_x = stack.mean_filter(x,
                                   kernel_shape="square",
                                   kernel_size=3)
    expected_x = np.array(
        [[2, 1, 0, 0, 0],
         [1, 1, 0, 0, 0],
         [0, 0, 1, 0, 0],
         [0, 0, 1, 0, 0],
         [0, 0, 0, 0, 0]],
        dtype=np.uint8)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.uint8

    # np.uint16
    filtered_x = stack.mean_filter(x.astype(np.uint16),
                                   kernel_shape="square",
                                   kernel_size=3)
    expected_x = expected_x.astype(np.uint16)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.uint16

    # np.float32
    filtered_x = stack.mean_filter(x.astype(np.float32),
                                   kernel_shape="square",
                                   kernel_size=3)
    expected_x = np.array(
        [[2.333, 1.444, 0.556, 0., 0.],
         [1.556, 1., 0.444, 0., 0.],
         [0.889, 0.778, 1.111, 0.667, 0.556],
         [0.333, 0.444, 1., 0.667, 0.556],
         [0.222, 0.333, 0.889, 0.667, 0.556]],
        dtype=np.float32)
    assert_allclose(filtered_x, expected_x, rtol=1e-02)
    assert filtered_x.dtype == np.float32

    # np.float64
    filtered_x = stack.mean_filter(x.astype(np.float64),
                                   kernel_shape="square",
                                   kernel_size=3)
    expected_x = expected_x.astype(np.float64)
    assert_allclose(filtered_x, expected_x, rtol=1e-02)
    assert filtered_x.dtype == np.float64


def test_median_filter():
    # np.uint8
    filtered_x = stack.median_filter(x,
                                     kernel_shape="square",
                                     kernel_size=3)
    expected_x = np.array(
        [[2, 2, 0, 0, 0],
         [2, 1, 0, 0, 0],
         [1, 1, 1, 0, 0],
         [0, 0, 0, 0, 0],
         [0, 0, 1, 0, 0]],
        dtype=np.uint8)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.uint8

    # np.uint16
    filtered_x = stack.median_filter(x.astype(np.uint16),
                                     kernel_shape="square",
                                     kernel_size=3)
    expected_x = expected_x.astype(np.uint16)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.uint16


def test_maximum_filter():
    # np.uint8
    filtered_x = stack.maximum_filter(x,
                                      kernel_shape="square",
                                      kernel_size=3)
    expected_x = np.array(
        [[3, 3, 2, 0, 0],
         [3, 3, 2, 0, 0],
         [2, 2, 5, 5, 5],
         [2, 2, 5, 5, 5],
         [2, 2, 5, 5, 5]],
        dtype=np.uint8)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.uint8

    # np.uint16
    filtered_x = stack.maximum_filter(x.astype(np.uint16),
                                      kernel_shape="square",
                                      kernel_size=3)
    expected_x = expected_x.astype(np.uint16)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.uint16


def test_minimum_filter():
    # np.uint8
    filtered_x = stack.minimum_filter(x,
                                      kernel_shape="square",
                                      kernel_size=3)
    expected_x = np.array(
        [[1, 0, 0, 0, 0],
         [0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0]],
        dtype=np.uint8)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.uint8

    # np.uint16
    filtered_x = stack.minimum_filter(x.astype(np.uint16),
                                      kernel_shape="square",
                                      kernel_size=3)
    expected_x = expected_x.astype(np.uint16)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.uint16


def test_log_filter():
    # float64
    y_float64 = stack.cast_img_float64(y)
    filtered_y_float64 = stack.log_filter(y_float64, 2)
    expected_y_float64 = np.array(
        [[0., 0., 0.02995949, 0.06212277, 0.07584532],
         [0., 0., 0.02581818, 0.05134284, 0.06123539],
         [0., 0., 0.01196859, 0.0253716, 0.02853162],
         [0., 0., 0., 0., 0.],
         [0., 0., 0., 0., 0.]],
        dtype=np.float64)
    assert_allclose(filtered_y_float64, expected_y_float64, rtol=1e-6)
    assert filtered_y_float64.dtype == np.float64

    # float32
    y_float32 = stack.cast_img_float32(y)
    filtered_y = stack.log_filter(y_float32, 2)
    expected_y = stack.cast_img_float32(expected_y_float64)
    assert_allclose(filtered_y, expected_y, rtol=1e-6)
    assert filtered_y.dtype == np.float32

    # uint8
    filtered_y = stack.log_filter(y, 2)
    expected_y = stack.cast_img_uint8(expected_y_float64)
    assert_array_equal(filtered_y, expected_y)
    assert filtered_y.dtype == np.uint8

    # uint16
    y_uint16 = stack.cast_img_uint16(y)
    filtered_y = stack.log_filter(y_uint16, 2)
    expected_y = stack.cast_img_uint16(expected_y_float64)
    assert_array_equal(filtered_y, expected_y)
    assert filtered_y.dtype == np.uint16


def test_gaussian_filter():
    # float64
    y_float64 = stack.cast_img_float64(y)
    filtered_y_float64 = stack.gaussian_filter(y_float64, 2)
    expected_y_float64 = np.array(
        [[0.08928096, 0.1573019 , 0.22897881, 0.28086597, 0.3001061 ],
         [0.08668051, 0.14896399, 0.21282558, 0.25752308, 0.27253406],
         [0.07634613, 0.12664142, 0.17574502, 0.20765944, 0.2155001 ],
         [0.05890843, 0.09356377, 0.12493327, 0.1427122 , 0.14374558],
         [0.03878372, 0.05873308, 0.07492625, 0.08201409, 0.07939603]],
        dtype=np.float64)
    assert_allclose(filtered_y_float64, expected_y_float64, rtol=1e-6)
    assert filtered_y_float64.dtype == np.float64

    # float32
    y_float32 = stack.cast_img_float32(y)
    filtered_y = stack.gaussian_filter(y_float32, 2)
    expected_y = stack.cast_img_float32(expected_y_float64)
    assert_allclose(filtered_y, expected_y, rtol=1e-6)
    assert filtered_y.dtype == np.float32

    # uint8
    with pytest.raises(ValueError):
        stack.gaussian_filter(y, 2, allow_negative=True)
    filtered_y = stack.gaussian_filter(y, 2)
    expected_y = stack.cast_img_uint8(expected_y_float64)
    assert_array_equal(filtered_y, expected_y)
    assert filtered_y.dtype == np.uint8

    # uint16
    y_uint16 = stack.cast_img_uint16(y)
    with pytest.raises(ValueError):
        stack.gaussian_filter(y_uint16, 2, allow_negative=True)
    filtered_y = stack.gaussian_filter(y_uint16, 2)
    expected_y = stack.cast_img_uint16(expected_y_float64)
    assert_array_equal(filtered_y, expected_y)
    assert filtered_y.dtype == np.uint16


def test_background_removal_mean():
    # np.uint8
    filtered_x = stack.remove_background_mean(x,
                                              kernel_shape="square",
                                              kernel_size=3)
    expected_x = np.array(
        [[1, 1, 0, 0, 0],
         [1, 0, 0, 0, 0],
         [0, 1, 0, 0, 0],
         [0, 2, 0, 5, 0],
         [0, 0, 0, 0, 0]],
        dtype=np.uint8)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.uint8

    # np.uint16
    filtered_x = stack.remove_background_mean(x.astype(np.uint16),
                                              kernel_shape="square",
                                              kernel_size=3)
    expected_x = expected_x.astype(np.uint16)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.uint16


def test_background_removal_gaussian():
    # float64
    y_float64 = stack.cast_img_float64(y)
    filtered_y_float64 = stack.remove_background_gaussian(y_float64, 2)
    expected_y_float64 = np.array(
        [[0., 0., 0.01415845, 0.36227129, 0.],
         [0., 0., 0.25776265, 0.66404555, 0.43726986],
         [0., 0., 0.11052949, 0.59626213, 0.],
         [0., 0.42016172, 0., 0., 0.],
         [0., 0., 0., 0., 0.]],
        dtype=np.float64)
    assert_allclose(filtered_y_float64, expected_y_float64, rtol=1e-6)
    assert filtered_y_float64.dtype == np.float64

    # float32
    y_float32 = stack.cast_img_float32(y)
    filtered_y = stack.remove_background_gaussian(y_float32, 2)
    expected_y = stack.cast_img_float32(expected_y_float64)
    assert_allclose(filtered_y, expected_y, rtol=1e-6)
    assert filtered_y.dtype == np.float32

    # uint8
    with pytest.raises(ValueError):
        stack.gaussian_filter(y, 2, allow_negative=True)
    filtered_y = stack.remove_background_gaussian(y, 2)
    expected_y = stack.cast_img_uint8(expected_y_float64)
    assert_array_equal(filtered_y, expected_y)
    assert filtered_y.dtype == np.uint8

    # uint16
    y_uint16 = stack.cast_img_uint16(y)
    with pytest.raises(ValueError):
        stack.gaussian_filter(y_uint16, 2, allow_negative=True)
    filtered_y = stack.remove_background_gaussian(y_uint16, 2)
    expected_y = stack.cast_img_uint16(expected_y_float64)
    assert_array_equal(filtered_y, expected_y)
    assert filtered_y.dtype == np.uint16


def test_dilation_filter():
    # np.uint8
    filtered_x = stack.dilation_filter(x,
                                       kernel_shape="square",
                                       kernel_size=3)
    expected_x = np.array(
        [[3, 3, 2, 0, 0],
         [3, 3, 2, 0, 0],
         [2, 2, 5, 5, 5],
         [2, 2, 5, 5, 5],
         [2, 2, 5, 5, 5]],
        dtype=np.uint8)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.uint8

    # np.uint16
    filtered_x = stack.dilation_filter(x.astype(np.uint16),
                                       kernel_shape="square",
                                       kernel_size=3)
    expected_x = expected_x.astype(np.uint16)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.uint16

    # np.float32
    filtered_x = stack.dilation_filter(x.astype(np.float32),
                                       kernel_shape="square",
                                       kernel_size=3)
    expected_x = expected_x.astype(np.float32)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.float32

    # np.float64
    filtered_x = stack.dilation_filter(x.astype(np.float64),
                                       kernel_shape="square",
                                       kernel_size=3)
    expected_x = expected_x.astype(np.float64)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.float64

    # bool
    filtered_x = stack.dilation_filter(x.astype(bool),
                                       kernel_shape="square",
                                       kernel_size=3)
    expected_x = expected_x.astype(bool)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == bool


def test_erosion_filter():
    # np.uint8
    filtered_x = stack.erosion_filter(x,
                                      kernel_shape="square",
                                      kernel_size=3)
    expected_x = np.array(
        [[1, 0, 0, 0, 0],
         [0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0]],
        dtype=np.uint8)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.uint8

    # np.uint16
    filtered_x = stack.erosion_filter(x.astype(np.uint16),
                                      kernel_shape="square",
                                      kernel_size=3)
    expected_x = expected_x.astype(np.uint16)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.uint16

    # np.float32
    filtered_x = stack.erosion_filter(x.astype(np.float32),
                                      kernel_shape="square",
                                      kernel_size=3)
    expected_x = expected_x.astype(np.float32)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.float32

    # np.float64
    filtered_x = stack.erosion_filter(x.astype(np.float64),
                                      kernel_shape="square",
                                      kernel_size=3)
    expected_x = expected_x.astype(np.float64)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == np.float64

    # bool
    filtered_x = stack.erosion_filter(x.astype(bool),
                                      kernel_shape="square",
                                      kernel_size=3)
    expected_x = expected_x.astype(bool)
    assert_array_equal(filtered_x, expected_x)
    assert filtered_x.dtype == bool



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
The bigfish.classification subpackage includes functions to prepare input data,
craft features and train classification models.
"""

from .input_preparation import prepare_extracted_data

from .features import compute_features
from .features import get_features_name
from .features import features_distance
from .features import features_in_out_nucleus
from .features import features_protrusion
from .features import features_dispersion
from .features import features_topography
from .features import features_foci
from .features import features_area
from .features import features_centrosome


_input_preparation = [
    "prepare_extracted_data"]

_features = [
    "compute_features",
    "get_features_name",
    "features_distance"
    "features_in_out_nucleus"
    "features_protrusion"
    "features_dispersion"
    "features_topography"
    "features_foci"
    "features_area"
    "features_centrosome"]

__all__ = _input_preparation + _features



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Functions to detect spots in 2-d and 3-d.
"""

import warnings
import scipy.ndimage as ndi
import numpy as np

import bigfish.stack as stack

from skimage.measure import regionprops
from skimage.measure import label


# ### Main function ###

def detect_spots(images, threshold=None, remove_duplicate=True,
                 return_threshold=False, voxel_size_z=None, voxel_size_yx=100,
                 psf_z=None, psf_yx=200):
    """Apply LoG filter followed by a Local Maximum algorithm to detect spots
    in a 2-d or 3-d image.

    #. We smooth the image with a LoG filter.
    #. We apply a multidimensional maximum filter.
    #. A pixel which has the same value in the original and filtered images
       is a local maximum.
    #. We remove local peaks under a threshold.
    #. We keep only one pixel coordinate per detected spot.

    Parameters
    ----------
    images : List[np.ndarray] or np.ndarray
        Image (or list of images) with shape (z, y, x) or (y, x). If several
        images are provided, the same threshold is applied.
    threshold : float or int
        A threshold to discriminate relevant spots from noisy blobs. If None,
        optimal threshold is selected automatically. If several images are
        provided, one optimal threshold is selected for all the images.
    remove_duplicate : bool
        Remove potential duplicate coordinates for the same spots. Slow the
        running.
    return_threshold : bool
        Return the threshold used to detect spots.
    voxel_size_z : int or float or None
        Height of a voxel, along the z axis, in nanometer. If None, image is
        considered in 2-d.
    voxel_size_yx : int or float
        Size of a voxel on the yx plan, in nanometer.
    psf_z : int or float or None
        Theoretical size of the PSF emitted by a spot in the z plan,
        in nanometer. If None, image is considered in 2-d.
    psf_yx : int or float
        Theoretical size of the PSF emitted by a spot in the yx plan,
        in nanometer.

    Returns
    -------
    spots : List[np.ndarray] or np.ndarray, np.int64
        Coordinates (or list of coordinates) of the spots with shape
        (nb_spots, 3) or (nb_spots, 2), for 3-d or 2-d images respectively.
    threshold : int or float
        Threshold used to discriminate spots from noisy blobs.

    """
    # check parameters
    stack.check_parameter(threshold=(float, int, type(None)),
                          remove_duplicate=bool,
                          return_threshold=bool,
                          voxel_size_z=(int, float, type(None)),
                          voxel_size_yx=(int, float),
                          psf_z=(int, float, type(None)),
                          psf_yx=(int, float))

    # if one image is provided we enlist it
    if not isinstance(images, list):
        stack.check_array(images,
                          ndim=[2, 3],
                          dtype=[np.uint8, np.uint16,
                                 np.float32, np.float64])
        ndim = images.ndim
        images = [images]
        is_list = False
    else:
        ndim = None
        for i, image in enumerate(images):
            stack.check_array(image,
                              ndim=[2, 3],
                              dtype=[np.uint8, np.uint16,
                                     np.float32, np.float64])
            if i == 0:
                ndim = image.ndim
            else:
                if ndim != image.ndim:
                    raise ValueError("Provided images should have the same "
                                     "number of dimensions.")
        is_list = True

    # check consistency between parameters
    if ndim == 3 and voxel_size_z is None:
        raise ValueError("Provided images has {0} dimensions but "
                         "'voxel_size_z' parameter is missing.".format(ndim))
    if ndim == 3 and psf_z is None:
        raise ValueError("Provided images has {0} dimensions but "
                         "'psf_z' parameter is missing.".format(ndim))
    if ndim == 2:
        voxel_size_z = None
        psf_z = None

    # detect spots
    if return_threshold:
        spots, threshold = _detect_spots_from_images(
            images,
            threshold=threshold,
            remove_duplicate=remove_duplicate,
            return_threshold=return_threshold,
            voxel_size_z=voxel_size_z,
            voxel_size_yx=voxel_size_yx,
            psf_z=psf_z,
            psf_yx=psf_yx)
    else:
        spots = _detect_spots_from_images(
            images,
            threshold=threshold,
            remove_duplicate=remove_duplicate,
            return_threshold=return_threshold,
            voxel_size_z=voxel_size_z,
            voxel_size_yx=voxel_size_yx,
            psf_z=psf_z,
            psf_yx=psf_yx)

    # format results
    if not is_list:
        spots = spots[0]

    # return threshold or not
    if return_threshold:
        return spots, threshold
    else:
        return spots


def _detect_spots_from_images(images, threshold=None, remove_duplicate=True,
                              return_threshold=False, voxel_size_z=None,
                              voxel_size_yx=100, psf_z=None, psf_yx=200):
    """Apply LoG filter followed by a Local Maximum algorithm to detect spots
    in a 2-d or 3-d image.

    #. We smooth the image with a LoG filter.
    #. We apply a multidimensional maximum filter.
    #. A pixel which has the same value in the original and filtered images
       is a local maximum.
    #. We remove local peaks under a threshold.
    #. We keep only one pixel coordinate per detected spot.

    Parameters
    ----------
    images : List[np.ndarray]
        List of images with shape (z, y, x) or (y, x). The same threshold is
        applied to every images.
    threshold : float or int
        A threshold to discriminate relevant spots from noisy blobs. If None,
        optimal threshold is selected automatically. If several images are
        provided, one optimal threshold is selected for all the images.
    remove_duplicate : bool
        Remove potential duplicate coordinates for the same spots. Slow the
        running.
    return_threshold : bool
        Return the threshold used to detect spots.
    voxel_size_z : int or float or None
        Height of a voxel, along the z axis, in nanometer. If None, image is
        considered in 2-d.
    voxel_size_yx : int or float
        Size of a voxel on the yx plan, in nanometer.
    psf_z : int or float or None
        Theoretical size of the PSF emitted by a spot in the z plan, in
        nanometer. If None, image is considered in 2-d.
    psf_yx : int or float
        Theoretical size of the PSF emitted by a spot in the yx plan, in
        nanometer.

    Returns
    -------
    all_spots : List[np.ndarray], np.int64
        List of spot coordinates with shape (nb_spots, 3) or (nb_spots, 2),
        for 3-d or 2-d images respectively.
    threshold : int or float
        Threshold used to discriminate spots from noisy blobs.

    """
    # initialization
    sigma = stack.get_sigma(voxel_size_z, voxel_size_yx, psf_z, psf_yx)
    n = len(images)

    # apply LoG filter and find local maximum
    images_filtered = []
    pixel_values = []
    masks = []
    for image in images:
        # filter image
        image_filtered = stack.log_filter(image, sigma)
        images_filtered.append(image_filtered)

        # get pixels value
        pixel_values += list(image_filtered.ravel())

        # find local maximum
        mask_local_max = local_maximum_detection(image_filtered, sigma)
        masks.append(mask_local_max)

    # get optimal threshold if necessary based on all the images
    if threshold is None:

        # get threshold values we want to test
        thresholds = _get_candidate_thresholds(pixel_values)

        # get spots count and its logarithm
        all_value_spots = []
        minimum_threshold = float(thresholds[0])
        for i in range(n):
            image_filtered = images_filtered[i]
            mask_local_max = masks[i]
            spots, mask_spots = spots_thresholding(
                image_filtered, mask_local_max,
                threshold=minimum_threshold,
                remove_duplicate=False)
            value_spots = image_filtered[mask_spots]
            all_value_spots.append(value_spots)
        all_value_spots = np.concatenate(all_value_spots)
        thresholds, count_spots = _get_spot_counts(thresholds, all_value_spots)

        # select threshold where the kink of the distribution is located
        if count_spots.size > 0:
            threshold, _, _ = _get_breaking_point(thresholds, count_spots)

    # detect spots
    all_spots = []
    for i in range(n):

        # get images and masks
        image_filtered = images_filtered[i]
        mask_local_max = masks[i]

        # detection
        spots, _ = spots_thresholding(image_filtered, mask_local_max,
                                      threshold, remove_duplicate)
        all_spots.append(spots)

    # return threshold or not
    if return_threshold:
        return all_spots, threshold
    else:
        return all_spots


# ### LoG spot detection ###

def local_maximum_detection(image, min_distance):
    """Compute a mask to keep only local maximum, in 2-d and 3-d.

    #. We apply a multidimensional maximum filter.
    #. A pixel which has the same value in the original and filtered images
       is a local maximum.

    Several connected pixels can have the same value. In such a case, the
    local maximum is not unique.

    In order to make the detection robust, it should be applied to a
    filtered image (using :func:`bigfish.stack.log_filter` for example).

    Parameters
    ----------
    image : np.ndarray
        Image to process with shape (z, y, x) or (y, x).
    min_distance : int, float or Tuple(float)
        Minimum distance (in pixels) between two spots we want to be able to
        detect separately. One value per spatial dimension (zyx or yx
        dimensions). If it's a scalar, the same distance is applied to every
        dimensions.

    Returns
    -------
    mask : np.ndarray, bool
        Mask with shape (z, y, x) or (y, x) indicating the local peaks.

    """
    # check parameters
    stack.check_array(image,
                      ndim=[2, 3],
                      dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_parameter(min_distance=(float, int, tuple))

    # compute the kernel size (centered around our pixel because it is uneven)
    if isinstance(min_distance, (int, float)):
        min_distance = (min_distance,) * image.ndim
        min_distance = np.ceil(min_distance).astype(image.dtype)
    elif image.ndim != len(min_distance):
        raise ValueError("'min_distance' should be a scalar or a tuple with "
                         "one value per dimension. Here the image has {0} "
                         "dimensions and 'min_distance' {1} elements."
                         .format(image.ndim, len(min_distance)))
    else:
        min_distance = np.ceil(min_distance).astype(image.dtype)
    kernel_size = 2 * min_distance + 1

    # apply maximum filter to the original image
    image_filtered = ndi.maximum_filter(image, size=kernel_size)

    # we keep the pixels with the same value before and after the filtering
    mask = image == image_filtered

    return mask


def spots_thresholding(image, mask_local_max, threshold,
                       remove_duplicate=True):
    """Filter detected spots and get coordinates of the remaining spots.

    In order to make the thresholding robust, it should be applied to a
    filtered image (using :func:`bigfish.stack.log_filter`
    for example). If the local maximum is not unique (it can happen if connected
    pixels have the same value), a connected component algorithm is applied to
    keep only one coordinate per spot.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x) or (y, x).
    mask_local_max : np.ndarray, bool
        Mask with shape (z, y, x) or (y, x) indicating the local peaks.
    threshold : float, int or None
        A threshold to discriminate relevant spots from noisy blobs. If None,
        detection is aborted with a warning.
    remove_duplicate : bool
        Remove potential duplicate coordinates for the same spots. Slow the
        running.

    Returns
    -------
    spots : np.ndarray, np.int64
        Coordinate of the local peaks with shape (nb_peaks, 3) or
        (nb_peaks, 2) for 3-d or 2-d images respectively.
    mask : np.ndarray, bool
        Mask with shape (z, y, x) or (y, x) indicating the spots.

    """
    # check parameters
    stack.check_array(image,
                      ndim=[2, 3],
                      dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_array(mask_local_max,
                      ndim=[2, 3],
                      dtype=[bool])
    stack.check_parameter(threshold=(float, int, type(None)),
                          remove_duplicate=bool)

    if threshold is None:
        mask = np.zeros_like(image, dtype=bool)
        spots = np.array([], dtype=np.int64).reshape((0, image.ndim))
        warnings.warn("No spots were detected (threshold is {0})."
                      .format(threshold),
                      UserWarning)
        return spots, mask

    # remove peak with a low intensity
    mask = (mask_local_max & (image > threshold))
    if mask.sum() == 0:
        spots = np.array([], dtype=np.int64).reshape((0, image.ndim))
        return spots, mask

    # make sure we detect only one coordinate per spot
    if remove_duplicate:
        # when several pixels are assigned to the same spot, keep the centroid
        cc = label(mask)
        local_max_regions = regionprops(cc)
        spots = []
        for local_max_region in local_max_regions:
            spot = np.array(local_max_region.centroid)
            spots.append(spot)
        spots = np.stack(spots).astype(np.int64)

        # built mask again
        mask = np.zeros_like(mask)
        mask[spots[:, 0], spots[:, 1]] = True

    else:
        # get peak coordinates
        spots = np.nonzero(mask)
        spots = np.column_stack(spots)

    # case where no spots were detected
    if spots.size == 0:
        warnings.warn("No spots were detected (threshold is {0})."
                      .format(threshold),
                      UserWarning)

    return spots, mask


# ### Threshold selection ###

def automated_threshold_setting(image, mask_local_max):
    """Automatically set the optimal threshold to detect spots.

    In order to make the thresholding robust, it should be applied to a
    filtered image (using :func:`bigfish.stack.log_filter` for example). The
    optimal threshold is selected based on the spots distribution. The latter
    should have an elbow curve discriminating a fast decreasing stage from a
    more stable one (a plateau).

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x) or (y, x).
    mask_local_max : np.ndarray, bool
        Mask with shape (z, y, x) or (y, x) indicating the local peaks.

    Returns
    -------
    optimal_threshold : int
        Optimal threshold to discriminate spots from noisy blobs.

    """
    # check parameters
    stack.check_array(image,
                      ndim=[2, 3],
                      dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_array(mask_local_max,
                      ndim=[2, 3],
                      dtype=[bool])

    # get threshold values we want to test
    thresholds = _get_candidate_thresholds(image.ravel())

    # get spots count and its logarithm
    first_threshold = float(thresholds[0])
    spots, mask_spots = spots_thresholding(
        image, mask_local_max, first_threshold, remove_duplicate=False)
    value_spots = image[mask_spots]
    thresholds, count_spots = _get_spot_counts(thresholds, value_spots)

    # select threshold where the break of the distribution is located
    if count_spots.size > 0:
        optimal_threshold, _, _ = _get_breaking_point(thresholds, count_spots)

    # case where no spots were detected
    else:
        optimal_threshold = None

    return optimal_threshold


def _get_candidate_thresholds(pixel_values):
    """Choose the candidate thresholds to test for the spot detection.

    Parameters
    ----------
    pixel_values : np.ndarray
        Pixel intensity values of the image.

    Returns
    -------
    thresholds : np.ndarray, np.float64
        Candidate threshold values.

    """
    # choose appropriate thresholds candidate
    start_range = 0
    end_range = int(np.percentile(pixel_values, 99.9999))
    if end_range < 100:
        thresholds = np.linspace(start_range, end_range, num=100)
    else:
        thresholds = [i for i in range(start_range, end_range + 1)]
    thresholds = np.array(thresholds)

    return thresholds


def _get_spot_counts(thresholds, value_spots):
    """Compute and format the spots count function for different thresholds.

    Parameters
    ----------
    thresholds : np.ndarray, np.float64
        Candidate threshold values.
    value_spots : np.ndarray
        Pixel intensity values of all spots.

    Returns
    -------
    thresholds : np.ndarray, np.float64
        Candidate threshold values.
    count_spots : np.ndarray, np.float64
        Spots count function.

    """
    # count spots for each threshold
    count_spots = np.log([np.count_nonzero(value_spots > t)
                          for t in thresholds])
    count_spots = stack.centered_moving_average(count_spots, n=5)

    # the tail of the curve unnecessarily flatten the slop
    count_spots = count_spots[count_spots > 2]
    thresholds = thresholds[:count_spots.size]

    return thresholds, count_spots


def _get_breaking_point(x, y):
    """Select the x-axis value where a L-curve has a kink.

    Assuming a L-curve from A to B, the 'breaking_point' is the more distant
    point to the segment [A, B].

    Parameters
    ----------
    x : np.array, np.float64
        X-axis values.
    y : np.array, np.float64
        Y-axis values.

    Returns
    -------
    breaking_point : float
        X-axis value at the kink location.
    x : np.array, np.float64
        X-axis values.
    y : np.array, np.float64
        Y-axis values.

    """
    # select threshold where curve break
    slope = (y[-1] - y[0]) / len(y)
    y_grad = np.gradient(y)
    m = list(y_grad >= slope)
    j = m.index(False)
    m = m[j:]
    x = x[j:]
    y = y[j:]
    if True in m:
        i = m.index(True)
    else:
        i = -1
    breaking_point = float(x[i])

    return breaking_point, x, y


def get_elbow_values(images, voxel_size_z=None, voxel_size_yx=100, psf_z=None,
                     psf_yx=200):
    """Get values to plot the elbow curve used to automatically set the
    threshold.

    Parameters
    ----------
    images : List[np.ndarray] or np.ndarray
        Image (or list of images) with shape (z, y, x) or (y, x). If several
        images are provided, the same threshold is applied.
    voxel_size_z : int or float or None
        Height of a voxel, along the z axis, in nanometer. If None, image is
        considered in 2-d.
    voxel_size_yx : int or float
        Size of a voxel on the yx plan, in nanometer.
    psf_z : int or float or None
        Theoretical size of the PSF emitted by a spot in the z plan,
        in nanometer. If None, image is considered in 2-d.
    psf_yx : int or float
        Theoretical size of the PSF emitted by a spot in the yx plan,
        in nanometer.

    Returns
    -------
    thresholds : np.ndarray, np.float64
        Candidate threshold values.
    count_spots : np.ndarray, np.float64
        Spots count function.
    threshold : float or None
        Threshold automatically set.

    """
    # check parameters
    stack.check_parameter(voxel_size_z=(int, float, type(None)),
                          voxel_size_yx=(int, float),
                          psf_z=(int, float, type(None)),
                          psf_yx=(int, float))

    # if one image is provided we enlist it
    if not isinstance(images, list):
        stack.check_array(images,
                          ndim=[2, 3],
                          dtype=[np.uint8, np.uint16,
                                 np.float32, np.float64])
        ndim = images.ndim
        images = [images]
        n = 1
    else:
        ndim = None
        for i, image in enumerate(images):
            stack.check_array(image,
                              ndim=[2, 3],
                              dtype=[np.uint8, np.uint16,
                                     np.float32, np.float64])
            if i == 0:
                ndim = image.ndim
            else:
                if ndim != image.ndim:
                    raise ValueError("Provided images should have the same "
                                     "number of dimensions.")
        n = len(images)

    # check consistency between parameters
    if ndim == 3 and voxel_size_z is None:
        raise ValueError("Provided images has {0} dimensions but "
                         "'voxel_size_z' parameter is missing.".format(ndim))
    if ndim == 3 and psf_z is None:
        raise ValueError("Provided images has {0} dimensions but "
                         "'psf_z' parameter is missing.".format(ndim))
    if ndim == 2:
        voxel_size_z = None
        psf_z = None

    # compute sigma
    sigma = stack.get_sigma(voxel_size_z, voxel_size_yx, psf_z, psf_yx)

    # apply LoG filter and find local maximum
    images_filtered = []
    pixel_values = []
    masks = []
    for image in images:
        # filter image
        image_filtered = stack.log_filter(image, sigma)
        images_filtered.append(image_filtered)

        # get pixels value
        pixel_values += list(image_filtered.ravel())

        # find local maximum
        mask_local_max = local_maximum_detection(
            image_filtered, sigma)
        masks.append(mask_local_max)

    # get threshold values we want to test
    thresholds = _get_candidate_thresholds(pixel_values)

    # get spots count and its logarithm
    all_value_spots = []
    minimum_threshold = float(thresholds[0])
    for i in range(n):
        image_filtered = images_filtered[i]
        mask_local_max = masks[i]
        spots, mask_spots = spots_thresholding(
            image_filtered, mask_local_max,
            threshold=minimum_threshold,
            remove_duplicate=False)
        value_spots = image_filtered[mask_spots]
        all_value_spots.append(value_spots)
    all_value_spots = np.concatenate(all_value_spots)
    thresholds, count_spots = _get_spot_counts(
        thresholds, all_value_spots)

    # select threshold where the kink of the distribution is located
    if count_spots.size > 0:
        threshold, _, _ = _get_breaking_point(thresholds, count_spots)
    else:
        threshold = None

    return thresholds, count_spots, threshold



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Functions to augment the data (images or coordinates).
"""

import numpy as np
from .preprocess import check_parameter
from .preprocess import check_array


def augment_2d(image):
    """Augment an image applying a random operation.

    Parameters
    ----------
    image : np.ndarray
        Image to augment with shape (y, x, channels) or (y, x, channels).

    Returns
    -------
    image_augmented : np.ndarray
        Image augmented with shape (y, x, channels).

    """
    # check input image
    check_parameter(image=np.ndarray)
    check_array(image, ndim=[2, 3])

    # randomly choose an operator
    operations = [_identity,
                  _flip_h, _flip_v,
                  _transpose, _transpose_inverse,
                  _rotation_90, _rotation_180, _rotation_270]
    random_operation = np.random.choice(operations)

    # augment the image
    image_augmented = random_operation(image)

    return image_augmented


def augment_2d_function(identity=False):
    """Choose a random operation to augment a 2-d image.

    Parameters
    ----------
    identity : bool
        Return identity function instead of a random transformation.

    Returns
    -------
    random_operation : callable
        Function to transform a 2-d image.

    """
    # no random transformation
    if identity:
        return _identity

    # randomly choose an operator
    operations = [_identity,
                  _flip_h, _flip_v,
                  _transpose, _transpose_inverse,
                  _rotation_90, _rotation_180, _rotation_270]
    random_operation = np.random.choice(operations)

    return random_operation


def augment_8_times(image):
    """Apply every transformation to a 2-d image.

    Parameters
    ----------
    image : np.ndarray
        Image to augment with shape (y, x, channels).

    Returns
    -------
    images_augmented : List[np.ndarray]
        List of images augmented with shape (y, x, channels).

    """
    # check input image
    check_parameter(image=np.ndarray)
    check_array(image, ndim=[2, 3])

    # initialization
    images_augmented = []

    # apply all operators
    operations = [_identity,
                  _flip_h, _flip_v,
                  _transpose, _transpose_inverse,
                  _rotation_90, _rotation_180, _rotation_270]
    for operation in operations:
        augmented_image = operation(image)
        images_augmented.append(augmented_image)

    return images_augmented


def augment_8_times_reversed(images_augmented):
    """Apply every transformation back to return the original 2-d image.

    Parameters
    ----------
    images_augmented : List[np.ndarray]
        List of images augmented with shape (y, x, channels).

    Returns
    -------
    images_original : List[np.ndarray]
        List of original images with shape (y, x, channels).

    """
    # check input image
    check_parameter(images_augmented=list)
    for image_augmented in images_augmented:
        check_array(image_augmented, ndim=[2, 3])

    # initialization
    images_original = []

    # apply all operators
    operations = [_identity,
                  _flip_h, _flip_v,
                  _transpose, _transpose_inverse,
                  _rotation_270, _rotation_180, _rotation_90]
    for i, image_augmented in enumerate(images_augmented):
        operation = operations[i]
        image_original = operation(image_augmented)
        images_original.append(image_original)

    return images_original


def _identity(image):
    """Do not apply any operation to the image.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (x, y, channels).

    Returns
    -------
    image : np.ndarray
        Image with shape (x, y, channels).

    """
    return image


def _flip_h(image):
    """Flip an image horizontally.

    Parameters
    ----------
    image : np.ndarray
        Image to flip with shape (y, x, channels).

    Returns
    -------
    image_flipped : np.ndarray
        Image flipped with shape (y, x, channels).

    """
    image_flipped = np.flip(image, axis=0)

    return image_flipped


def _flip_v(image):
    """Flip an image vertically.

    Parameters
    ----------
    image : np.ndarray
        Image to flip with shape (y, x, channels).

    Returns
    -------
    image_flipped : np.ndarray
        Image flipped with shape (y, x, channels).

    """
    image_flipped = np.flip(image, axis=1)

    return image_flipped


def _transpose(image):
    """Transpose an image (flip it along the top left - bottom right diagonal).

    Parameters
    ----------
    image : np.ndarray
        Image to transpose with shape (y, x, channels).

    Returns
    -------
    image_transposed : np.ndarray
        Image transposed with shape (y, x, channels).

    """
    if image.ndim == 3:
        image_transposed = np.transpose(image, axes=(1, 0, 2))
    else:
        image_transposed = np.transpose(image)

    return image_transposed


def _rotation_90(image):
    """Rotate an image with 90 degrees (clockwise).

    Parameters
    ----------
    image : np.ndarray
        Image to rotate with shape (y, x, channels).

    Returns
    -------
    image_rotated : np.ndarray
        Image rotated with shape (y, x, channels).

    """
    image_rotated = _flip_h(image)
    image_rotated = _transpose(image_rotated)

    return image_rotated


def _rotation_180(image):
    """Rotate an image with 180 degrees.

    Parameters
    ----------
    image : np.ndarray
        Image to rotate with shape (y, x, channels).

    Returns
    -------
    image_rotated : np.ndarray
        Image rotated with shape (y, x, channels).

    """
    image_rotated = _flip_v(image)
    image_rotated = _flip_h(image_rotated)

    return image_rotated


def _rotation_270(image):
    """Rotate an image with 270 degrees (clockwise).

    Parameters
    ----------
    image : np.ndarray
        Image to rotate with shape (y, x, channels).

    Returns
    -------
    image_rotated : np.ndarray
        Image rotated with shape (y, x, channels).

    """
    image_rotated = _flip_v(image)
    image_rotated = _transpose(image_rotated)

    return image_rotated


def _transpose_inverse(image):
    """Flip an image along the bottom left - top right diagonal.

    Parameters
    ----------
    image : np.ndarray
        Image to transpose with shape (y, x, channels).

    Returns
    -------
    image_transposed : np.ndarray
        Image transposed with shape (y, x, channels).

    """
    image_transposed = _transpose(image)
    image_transposed = _rotation_180(image_transposed)

    return image_transposed



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for bigfish.detection.dense_decomposition module.
"""

# TODO test bigfish.detection.decompose_dense
# TODO test bigfish.detection.get_dense_region
# TODO test bigfish.detection.simulate_gaussian_mixture



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Utility functions for bigfish.plot subpackage.
"""

import matplotlib.pyplot as plt
import numpy as np


def save_plot(path_output, ext):
    """Save the plot.

    Parameters
    ----------
    path_output : str
        Path to save the image (without extension).
    ext : str or List[str]
        Extension used to save the plot. If it is a list of strings, the plot
        will be saved several times.

    """
    # add extension at the end of the filename
    extension = "." + ext
    if extension not in path_output:
        path_output += extension

    # save the plot
    if isinstance(ext, str):
        # add extension at the end of the filename
        extension = "." + ext
        if extension not in path_output:
            path_output += extension
        plt.savefig(path_output, format=ext)
    elif isinstance(ext, list):
        for ext_ in ext:
            # add extension at the end of the filename
            extension = "." + ext_
            if extension not in path_output:
                path_output += extension
            plt.savefig(path_output, format=ext_)
    else:
        Warning("Plot is not saved because the extension is not valid: "
                "{0}.".format(ext))


def get_minmax_values(tensor):
    """Get the minimum and maximum value of the image according to its dtype.

    Parameters
    ----------
    tensor : np.ndarray
        A 2-d, 3-d or 5-d tensor with shape (y, x), (z, y, x) or
        (r, c, z, y, x) respectively.

    Returns
    -------
    vmin : int
        Minimum value display in the plot.
    vmax : int
        Maximum value display in the plot.

    """
    vmin, vmax = None, None
    if tensor.dtype == np.uint8:
        vmin, vmax = 0, 255
    elif tensor.dtype == np.uint16:
        vmin, vmax = 0, 65535
    elif tensor.dtype == np.float32:
        vmin, vmax = 0, 1
    elif tensor.dtype == np.float64:
        vmin, vmax = 0, 1
    elif tensor.dtype == bool:
        vmin, vmax = 0, 1

    return vmin, vmax


def create_colormap():
    """Create a shuffled colormap to display segmentation masks.

    Returns
    -------
    colormap : ListedColormap object
        Colormap for matplotlib.

    """
    import matplotlib.pyplot as plt

    values = np.linspace(0, 1, 256)
    np.random.shuffle(values)
    colormap = plt.cm.colors.ListedColormap(plt.cm.YlGnBu(values))
    colormap.set_bad(color='black')

    return colormap



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Setup script.
"""

from setuptools import setup, find_packages


# package description
DESCRIPTION = "Toolbox for the analysis of smFISH images."

# package version
VERSION = None
with open('bigfish/__init__.py', encoding='utf-8') as f:
    for row in f:
        if row.startswith('__version__'):
            VERSION = row.strip().split()[-1][1:-1]
            break

# package dependencies
with open("requirements.txt", encoding='utf-8') as f:
    REQUIREMENTS = [l.strip() for l in f.readlines() if l]
DEEPLEARNING_REQUIREMENTS = [
    'tensorflow == 2.3.0',
    'tensorflow-addons == 0.12.1']

# long description of the package
with open("README.md", "r") as f:
    LONG_DESCRIPTION = f.read()

# A list of classifiers to categorize the project (only used for searching and
# browsing projects on PyPI)
CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Science/Research',
    'Intended Audience :: Developers',
    'Topic :: Software Development',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Image Processing',
    'Topic :: Scientific/Engineering :: Image Recognition',
    'Operating System :: Unix',
    'Operating System :: MacOS',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.6',
    'License :: OSI Approved :: BSD License']

# setup
setup(name='big-fish',
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type="text/markdown",
      author='Arthur Imbert',
      author_email='arthur.imbert.pro@gmail.com',
      url='https://github.com/fish-quant/big-fish',
      packages=find_packages(),
      license='BSD 3-Clause License',
      python_requires='>=3.6',
      install_requires=REQUIREMENTS,
      extras_require={'deeplearning': DEEPLEARNING_REQUIREMENTS},
      classifiers=CLASSIFIERS)
